# -*- coding: utf-8 -*-
"""
##############################################################################

Default parameters grid for GridSearch (param_grid) and RandomizedSearch (distribution) using shallow models
Model available: 'svm', 'knn', 'sgd', 'lr','rf', 'gnb', 'nn','gboosting'
The values for grid are stored in dictionary and can be accessed using
param = param_shallow()
param_grid = param[model.lower()]['param_grid']
param_grid = param[model.lower()]['distribution']
Authors: Ana Marta Sequeira

Date:12/2020

Email:

##############################################################################
"""
from sklearn.utils.fixes import loguniform
from scipy.stats import uniform, truncnorm, randint


# https://blog.usejournal.com/a-comparison-of-grid-search-and-randomized-search-using-scikit-learn-29823179bc85
# https://blog.usejournal.com/a-comparison-of-grid-search-and-randomized-search-using-scikit-learn-29823179bc85
# https://scikit-learn.org/stable/modules/grid_search.html#grid-search
# https://scikit-learn.org/stable/auto_examples/model_selection/plot_randomized_search.html#sphx-glr-auto-examples-model-selection-plot-randomized-search-py

# In contrast to GridSearchCV, not all parameter values are tried out, but rather a fixed number of parameter settings
# is sampled from the specified distributions. The number of parameter settings that are tried is given by n_iter.
# If all parameters are presented as a list, sampling without replacement is performed. If at least one parameter
# is given as a distribution, sampling with replacement is used.
# It is highly recommended to use continuous distributions for continuous parameters

# For continuous parameters, such as C above, it is important to specify a continuous distribution to take full
# advantage of the randomization. This way, increasing n_iter will always lead to a finer search./

# The randomized search and the grid search explore exactly the same space of parameters.
# The result in parameter settings is quite similar, while the run time for randomized search is drastically lower.

# Tuning the hyper-parameters of an estimator
# A search consists of:
# an estimator (regressor or classifier such as sklearn.svm.SVC()),
# a parameter space;
# a method for searching or sampling candidates;
# a cross-validation scheme; and
# a score function.

# GridSearchCV exhaustively considers all parameter combinations
# RandomizedSearchCV can sample a given number of candidates from a parameter space with a specified distribution.

# put early stopping in grid searchs? sgd, Gboosting, nn, RF


def param_shallow():
    param = {'svc':
                 {'param_grid':
                      [{'clf__C': [0.01, 0.1, 1.0, 10],
                        'clf__kernel': ['linear']},
                       {'clf__C': [0.01, 0.1, 1.0, 10],
                        'clf__kernel': ['rbf'],
                        'clf__gamma': ['scale', 0.001, 0.0001]}],
                  'distribution':
                      [{'clf__C': loguniform(1e-2, 1e1),
                        'clf__kernel': ['linear']},
                       {'clf__C': loguniform(1e-2, 1e1),
                        'clf__gamma': loguniform(1e-4, 1e1),  # np.power(10, np.arange(-4, 1, dtype=float)),
                        'clf__kernel': ['rbf']}]
                  },

             'linear_svc':
                 {'param_grid':
                      [{'clf__C': [0.01, 0.1, 1.0, 10]}],
                  'distribution':
                      [{'clf__C': loguniform(1e-4, 1e1)}]
                  },

             'rf':
                 {'param_grid':
                      [{'clf__n_estimators': [10, 100, 500],
                        'clf__max_features': ['sqrt', 'log2'],
                        'clf__bootstrap': [True],
                        'clf__criterion': ["gini"]}],
                  'distribution':
                      [{'clf__n_estimators': randint(1e1, 1e3),
                        'clf__max_features': ['sqrt', 'log2'],
                        'clf__bootstrap': [True],
                        'clf__criterion': ["gini"]}]
                  },

             'gboosting':
                 {'param_grid':
                      [{'clf__n_estimators': [10, 100, 500],
                        'clf__max_depth': [1, 3, 5, 10],
                        'clf__max_features': [0.6, 0.9]}],
                  'distribution':
                      [{'clf__learning_rate': loguniform(1e-3, 1e0),
                        'clf__n_estimators': randint(1e1, 1e3),
                        'clf__max_depth': randint(1e0, 2e1),
                        'clf__max_features': loguniform(5e-1, 1)}]
                  },

             'knn':
                 {'param_grid': [{'clf__n_neighbors': [2, 5, 10, 15],
                                  'clf__weights': ['uniform', 'distance'],
                                  'clf__leaf_size': [15, 30, 60]}],
                  'distribution':
                      [{'clf__n_neighbors': randint(2e0, 3e1),
                        'clf__weights': ['uniform', 'distance'],
                        'clf__leaf_size': loguniform(1e1, 1e2)}]
                  },

             'sgd':
                 {'param_grid':
                      [{'clf__loss': ['hinge', 'log', 'modified_huber', 'perceptron'],
                        'clf__alpha': [0.00001, 0.0001, 0.001, 0.01],
                        'clf__early_stopping': [True],
                        'clf__validation_fraction': [0.2],
                        'clf__n_iter_no_change': [30]}],
                  'distribution':
                      [{'clf__loss': ['hinge', 'log', 'modified_huber', 'perceptron'],
                        'clf__alpha': loguniform(1e-5, 1e-2),
                        'clf__early_stopping': [True],
                        'clf__validation_fraction': [0.2],
                        'clf__n_iter_no_change': [50]}]
                  },

             'lr':
                 {'param_grid':
                      [{'clf__C': [0.01, 0.1, 1.0, 10.0],
                        'clf__solver': ['liblinear', 'lbfgs', 'sag']}],
                  'distribution':
                      [{'clf__C': loguniform(1e-2, 1e1),
                        'clf__solver': ['liblinear', 'lbfgs', 'sag']}]
                  },

             'gnb':
                 {'param_grid':
                      [{'clf__var_smoothing': [1e-12, 1e-9, 1e-6]}],
                  'distribution':
                      [{'clf__var_smoothing': loguniform(1e-12, 1e-6)}]
                  },

             'nn':
                 {'param_grid':
                      [{'clf__activation': ['logistic', 'tanh', 'relu'],
                        'clf__alpha': [0.00001, 0.0001, 0.001],
                        'clf__learning_rate_init': [0.0001, 0.001, 0.01],
                        'clf__early_stopping': [True],
                        'clf__validation_fraction': [0.2],
                        'clf__n_iter_no_change': [50]}],
                  'distribution':
                      [{'clf__activation': ['logistic', 'tanh', 'relu'],
                        'clf__alpha': loguniform(1e-5, 1e-3),
                        'clf__learning_rate_init': loguniform(1e-4, 1e-2),
                        'clf__early_stopping': [True],
                        'clf__validation_fraction': [0.2],
                        'clf__n_iter_no_change': [50]}]  # put a quarter of the iterations by default
                  },

             # todo have added !!!!!!!
             'svr':
                 {'param_grid':
                      [{'clf__kernel': [ 'poly', 'rbf', 'sigmoid'], #'linear',
                        'clf__C': [ 0.1, 1.0, 10],
                        'clf__epsilon': [0.01, 0.1, 0.2, 1.0, 5.0],  # acceptable error margin(ϵ)
                        }],
                  'distribution':
                      [{'clf__kernel': ['linear', 'poly', 'rbf', 'sigmoid'],
                        'clf__C': loguniform(1e-2, 1e1),
                        'clf__epsilon': loguniform(1e-2, 1e2)
                        }]
                  },
             # 'linear_svr':
             #     {'param_grid':
             #          [{'clf__C': [0.01, 0.1, 1.0, 10],
             #            'clf__epsilon': [0.01, 0.1, 0.2, 1.0],  # acceptable error margin(ϵ)
             #            'clf__loss':['epsilon_insensitive', 'squared_epsilon_insensitive'],
             #            'clf__intercept_scaling':[0.1,1.0,5.0,10]
             #            }],
             #      'distribution':
             #          [{'clf__C': loguniform(1e-2, 1e1),
             #            'clf__epsilon': loguniform(1e-2, 1e2),
             #            'clf__loss': ['epsilon_insensitive', 'squared_epsilon_insensitive'],
             #            'clf__intercept_scaling':loguniform(1e-1, 1e1),
             #            }]
             #      },
             'linear_svr':
                 {'param_grid':
                      [{'clf__C': [0.01, 0.1, 1.0],
                        'clf__epsilon': [0.01, 0.1, 1.0],  # acceptable error margin(ϵ)
                        'clf__loss': ['epsilon_insensitive', 'squared_epsilon_insensitive'],
                        'clf__intercept_scaling': [0.1, 1.0]
                        }],
                  'distribution':
                      [{'clf__C': loguniform(1e-2, 1e1),
                        'clf__epsilon': loguniform(1e-2, 1e2),
                        'clf__loss': ['epsilon_insensitive', 'squared_epsilon_insensitive'],
                        'clf__intercept_scaling': loguniform(1e-1, 1e1),
                        }]
                  },
             # todo add min samples leaf/ split and like that?

             'rfr':
                 {'param_grid':
                      [{'clf__n_estimators': [10, 100, 300, 500],
                        'clf__max_features': ['auto', 'sqrt', 'log2'],
                        }],
                  'distribution':
                      [{'clf__n_estimators': randint(1e1, 5e3),
                        'clf__max_features': ['auto', 'sqrt', 'log2'],
                        }]
                  },
             'gboostingr':
                 {'param_grid':
                      [{'clf__n_estimators': [100, 400, 300, 500],
                        'clf__learning_rate': [0.05, 0.1, 0.5],
                        'clf__max_features': ['auto', 'sqrt', 'log2'],
                        }],
                  'distribution':
                      [{'clf__n_estimators': randint(1e2, 5e3),
                        'clf__learning_rate': loguniform(0.05, 0.5),
                        'clf__max_features': ['auto', 'sqrt', 'log2'],
                        }]
                  },
             'histgboostingr':
                 {'param_grid':
                      [{'clf__max_iter': [50, 100, 200, 300, 500],
                        'clf__learning_rate': [0.05, 0.1, 0.5],
                        'clf__loss': ['squared_error', 'absolute_error', 'poisson'],
                        }],
                  'distribution':
                      [{'clf__n_estimators': randint(5e1, 5e3),
                        'clf__learning_rate': loguniform(0.05, 0.5),
                        'clf__loss': ['squared_error', 'absolute_error', 'poisson'],
                        }]
                  },

             'adaboostr':
                 {'param_grid':
                      [{'clf__n_estimators': [500, 300, 100],
                        'clf__learning_rate': [0.5, 1, 1.5],
                        }],
                  'distribution':
                      [{'clf__n_estimators': randint(1e1, 5e3),
                        'clf__learning_rate': loguniform(0.05, 2),
                        }]
                  },

             'tweedier':
                 {'param_grid':
                      [{'clf__power': [1,2,3],#, 2, 3], # 0 is normal. other better models and distribution is not above zero 2 and 3 do not include the zero on the interval
                        # underlying target distribution according to the following table 0 normal 1 poisson 2 gamma 3 inverse gaussian
                        'clf__alpha': [0, 0.5, 1],
                        # regularization strength. alpha = 0 is equivalent to unpenalized GLMs
                        }],
                  'distribution':
                      [{'clf__power': [1, 2, 3],
                        'clf__max_features': loguniform(0, 1),
                        }]
                  },
             'knr':
                 {'param_grid':
                      [{'clf__n_neighbors': [2, 5, 10, 15],
                        'clf__weights': ['uniform', 'distance'],
                        'clf__leaf_size': [15, 30, 50],
                        }],
                  'distribution':
                      [{'clf__n_neighbors': randint(2e1, 15),
                        'clf__weights': ['uniform', 'distance'],
                        'clf__leaf_size':randint(15, 50),
                        }]
                  },

    # melhorar isto ou nem vale a pena estar aqui
             'gpr':
                 {'param_grid':
                      [{'clf__alpha': [1e-10,1e-5, 1e-3] # 1e-10 default

                        }],
                  'distribution':
                      [{'clf__alpha': loguniform(1e-11, 1e-5)
                        }]
                  },
             'sgdr':
                 {'param_grid':
                      [{'clf__loss': ['squared_loss', 'huber', 'squared_epsilon_insensitive'], # todo change squared loss to squared_error
                        'clf__alpha': [0.001, 0.0001, 0.00001], #default 0.0001
                        'clf__learning_rate': ['adaptive','optimal','invscaling'],  # acceptable error margin(ϵ)
    # early_stopping True
                        }],
                  'distribution':
                      [{'clf__loss': ['squared_error', 'huber', 'epsilon_insensitive', 'squared_epsilon_insensitive'],
                        'clf___alpha': loguniform(1e-5, 1e-2),
                        'clf__learning_rate':  ['adaptive','optimal','invscaling'],
                        }]
                  },

             }
    return param



# -*- coding: utf-8 -*-
"""
##############################################################################

File containing a class intend to facilitate deep learning
The functions are based on the package tensorflow and keras.

Authors: Ana Marta Sequeira

Date: 12/2020

Email:

##############################################################################
"""

import os
import logging
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import datetime
import time

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
logging.disable(logging.WARNING)
logging.getLogger("tensorflow").setLevel(logging.FATAL)
import tensorflow as tf
import tensorflow.keras

from tensorflow.keras.wrappers.scikit_learn import KerasClassifier
from tensorflow.keras.callbacks import ModelCheckpoint
from tensorflow.keras.utils import plot_model
from tensorflow.keras.callbacks import EarlyStopping, ReduceLROnPlateau
from sklearn.model_selection import StratifiedKFold
from sklearn.metrics import accuracy_score, f1_score, log_loss, classification_report, \
    multilabel_confusion_matrix
from sklearn.metrics import roc_auc_score
from sklearn.metrics import confusion_matrix
from sklearn.metrics import precision_recall_curve, precision_score, recall_score
from sklearn.metrics import make_scorer
from sklearn.metrics import matthews_corrcoef
from sklearn.preprocessing import label_binarize
from sklearn.model_selection import ShuffleSplit
from sklearn.metrics import mean_squared_error
from sklearn.metrics import mean_squared_error, r2_score
from tensorflow.keras.metrics import RootMeanSquaredError
from .dl_basic_models.basic_models_dnn import create_dnn_embedding, create_dnn_simple
from .dl_basic_models.basic_models_lstm import create_lstm_embedding, create_lstm_bilstm_simple
from .dl_basic_models.basic_models_cnn import create_cnn_1D, create_cnn_2D
from .dl_basic_models.basic_models_hybrid import create_cnn_lstm
from .utils import timer
from .parameters_deep import param_deep
from .plot_curves_ml_propythia import plot_roc_curve, plot_precision_recall_curve, \
    plot_validation_curve, plot_learning_curve, plot_summary_loss, plot_summary_accuracy
from .param_optimizer import ParamOptimizer
from sklearn.metrics import roc_auc_score
from sklearn.metrics import explained_variance_score, max_error, mean_absolute_error, mean_squared_error, \
    mean_squared_log_error, median_absolute_error, r2_score, mean_poisson_deviance, mean_gamma_deviance
from .utils import timer
from .parameters_shallow import param_shallow
from .param_optimizer import ParamOptimizer

tf.debugging.set_log_device_placement(True)
strategy = tf.distribute.MirroredStrategy()


class DeepML:
    """
    Class used for performing train and evaluation of different DL models.
    """

    def __init__(self,
                 x_train, y_train,
                 x_test, y_test, number_classes=2, problem_type='binary',
                 x_dval=None, y_dval=None,
                 model=None,
                 epochs=100, batch_size=512,
                 callbacks=None,
                 reduce_lr=True, early_stopping=True, checkpoint=True, tensorboard=False,
                 early_stopping_patience=30, reduce_lr_patience=50, reduce_lr_factor=0.2, reduce_lr_min=0.00001,
                 path='',
                 report_name=None,
                 verbose=1, validation_split=0.1,
                 shuffle=True, class_weights=None):

        """
        https://www.tensorflow.org/api_docs/python/tf/keras/Model
        :param x_train: dataset/array with features or encodings for training (x_train)
        :param y_train: labels for training (y_train)
        :param x_test: dataset/array with features or encodings for testing (x_test)
        :param y_test: labels for testing (y_test)
        :param number_classes: number of classes of the problem
        :param problem_type: problem type. 'binary', 'multiclass', 'multilabel'. Only used to run base models.
        Not necessary when running outside architectures.
        :param x_dval: dataset/array with features or encodings for vlidation (x_dval). If none,
        will retrive from train data using the param validation split.
        :param y_dval: labels for validation (y_dval)
        :param model: trained model. None by default.
        :param epochs: number of epochs to use when training DL model. default 100
        :param batch_size: batch size to use when trianing DL models. 512 default.
        :param callbacks: callbacks for training the model. If None (default), callbacks will be assigned.
        :param reduce_lr: if True (default),  Reduce learning rate will be added to callbacks
        :param early_stopping: If True (default) early stopping will be added to callbacks.
        :param checkpoint: If True (default) model checkpoint will be added to callbacks.
        :param tensorboard: If True (default) tensorboard extension will be added to callbacks. False by default
        :param early_stopping_patience: Early stopping parameter. number of epochs patience of no improvements
        to model stop. 30 by default.
        :param reduce_lr_patience: Reduce lr parameter. number of epochs with no improvement after which learning rate
        will be reduced 50 default.
        :param reduce_lr_factor: factor to reduce learning rate. 0.2 by default
        :param reduce_lr_min: lower bound on the learning rate. 0.00001 by default
        :param path: os path to store graphics and reports.
        :param report_name: if None will generate a txt file reporting results for the functions called inside class.
        :param verbose: verbose for model training . 1 default
        :param validation_split: fraction indicating the percentage to split train and validation. 0.1 by default.
        Only used if no validation (x_dval) is given.  The validation is retrieved from the last training examples.
        :param shuffle: whether to shuffle or not the data in traiing process. True by default.
        :param class_weights: class weights if to be used. None by default.
        """

        self.x_train = x_train
        self.x_test = x_test

        self.y_train = y_train
        self.y_test = y_test

        self.x_dval = x_dval
        self.y_dval = y_dval

        self.validation_split = validation_split

        self.callbacks = callbacks
        self.reduce_lr = reduce_lr
        self.early_stopping = early_stopping
        self.checkpoint = checkpoint
        self.early_stopping_patience = early_stopping_patience
        self.reduce_lr_patience = reduce_lr_patience
        self.reduce_lr_factor = reduce_lr_factor
        self.reduce_lr_min = reduce_lr_min
        self.tensorboard = tensorboard

        self.path = path

        self.model = model
        self.epochs = epochs
        self.verbose = verbose
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.class_weights = class_weights
        self.history = None

        self.final_units = number_classes

        self.problem_type = problem_type
        # self.final_units = len(y_train[0])
        # self.final_units = len(np.unique(y_train))
        # https://machinelearningmastery.com/how-to-choose-loss-functions-when-training-deep-learning-neural-networks/
        # https://machinelearningmastery.com/multi-label-classification-with-deep-learning/
        # https://towardsdatascience.com/journey-to-the-center-of-multi-label-classification-384c40229bff
        if problem_type == 'binary':  # binary problem ou se for multilabel
            self.loss = 'binary_crossentropy'
            self.activation = 'sigmoid'
            self.metrics = ['accuracy']
            self.final_units = 1
        elif problem_type == 'multiclass':  # multiclass sem ser multilabel
            # if any([item for sublist in self.y_train for item in sublist if item not in [0,1]]): # labeled # todo put here fastest way
            #     self.loss = 'sparse_categorical_crossentropy' # labeled encoder
            # if not hot encoded but labeled encoder
            # else:
            self.loss = 'sparse_categorical_crossentropy'
            self.activation = 'softmax'
            self.metrics = ['accuracy']
        elif problem_type == 'multilabel':
            self.loss = 'binary_crossentropy'
            self.activation = 'sigmoid'
            self.metrics = ['accuracy']
        elif problem_type == 'regression':  # this must be accepted otside
            self.loss = 'binary_crossentropy'  # mse
            self.activation = 'sigmoid'
            # self.metrics = [tf.keras.metrics.RootMeanSquaredError()]   #tf.keras.metrics.RootMeanSquaredError()
            self.final_units = 1
        # print(problem_type)
        # print(self.loss)
        # print(self.activation)
        # print(self.metrics)
        # print(self.final_units)
        # todo
        # adicionar regressao. e ver todas as coisas que têm metricas que possam ser alteradas

        self.report_name = report_name
        self.y_test_pred = None
        self.y_test_proba = None
        saved_args = locals()
        if self.report_name:
            self._report(str(self.report_name))
            self._report(saved_args)

    def _report(self, info, dataframe=False, **kwargs):
        filename = str(self.report_name)
        with open(filename, 'a+') as file:
            if dataframe is True:
                info.to_csv(file, sep='\t', mode='a', **kwargs)
            elif isinstance(info, str):
                file.writelines(info)
            else:
                for l in info:
                    file.writelines('\n{}'.format(l))

    # todo create methods to get all this values?

    def get_model(self):
        """
        Getter function for returning the keras compile model
        :return: keras compiled model
        """
        if self.model:
            return self.model
        else:
            raise ValueError("No model was compiled.")

    # ==================================================================================================================
    #  TRAIN AND RUN MODELS
    # ==================================================================================================================
    def train_model(self):
        """
        Function to train model
        :return: None
        """
        if self.x_dval is not None:
            validation_data = (self.x_dval, self.y_dval)
        else:
            validation_data = None

        self.history = self.model.fit(x=self.x_train, y=self.y_train,
                                      batch_size=self.batch_size, epochs=self.epochs, verbose=self.verbose,
                                      callbacks=self.callbacks,
                                      validation_split=self.validation_split, validation_data=validation_data,
                                      shuffle=self.shuffle, class_weight=self.class_weights,
                                      sample_weight=None, initial_epoch=0, steps_per_epoch=None,
                                      validation_steps=None, validation_batch_size=None, validation_freq=1,
                                      max_queue_size=10, workers=1, use_multiprocessing=False)

        # if data comes from generator
        # model.fit_generator(train_generator, samples_per_epoch=len(train_samples), validation_data=validation_generator,
        # nb_val_samples=len(validation_samples), nb_epoch=self.epochs)

        # # write report
        # if self.report_name is not None:
        #     self._report(['===TRAIN MODELS===\n', self.train_model.__name__, saved_args])

    def generate_callbacks(self):
        """
        Creates the callback instances to be used in the model training process
        :return: None
        """
        if self.callbacks is None:
            callbacks_list = []
            if self.early_stopping:
                # simple early stopping
                # es = EarlyStopping(monitor='val_loss', mode='min', verbose=1, patience=10)
                es = EarlyStopping(monitor='val_loss', mode='min', verbose=1,
                                   patience=self.early_stopping_patience)
                callbacks_list.append(es)

            if self.reduce_lr:
                # reduce learning rate reduce_lr_factor=0.5
                reduce_lr = ReduceLROnPlateau(monitor='val_loss', factor=self.reduce_lr_factor,
                                              patience=self.reduce_lr_patience, min_lr=self.reduce_lr_min, verbose=1)
                callbacks_list.append(reduce_lr)

            if self.checkpoint:
                # checkpoint
                filepath = os.path.join(self.path, 'weights.hdf5')
                # filepath = os.path.join(self.path, 'weights-{{epoch:02d}}-{{val_loss:.2f}}.hdf5')
                # filepath = str(self.path + '/weights-{{epoch:02d}}-{{val_loss:.2f}}.hdf5')
                cp = ModelCheckpoint(filepath=filepath, monitor='val_loss', verbose=1,
                                     save_best_only=True, save_weights_only=False,
                                     mode='auto', save_freq='epoch', period=1)
                callbacks_list.append(cp)

            if self.tensorboard:
                # TensorBoard is a tool for providing the measurements and visualizations needed during the machine
                # learning workflow. It enables tracking experiment metrics like loss and accuracy, visualizing the
                # model graph, projecting embeddings to a lower dimensional space, and much more.
                # https://www.tensorflow.org/tensorboard/get_started
                log_dir = "logs/fit/" + datetime.datetime.now().strftime("%Y%m%d-%H%M%S")
                tensorboard_callback = tf.keras.callbacks.TensorBoard(log_dir=log_dir, histogram_freq=1,
                                                                      write_graph=True,
                                                                      write_grads=False)
                callbacks_list.append(tensorboard_callback)
                # Start TensorBoard through the command line or within a notebook experience. The two interfaces are generally the same. In notebooks, use the %tensorboard line magic. On the command line, run the same command without "%".
                # %tensorboard --logdir logs/fit
            self.callbacks = callbacks_list
        # write report
        if self.report_name is not None:
            self._report(['===Callbacks===\n', self.generate_callbacks.__name__, self.callbacks])

    def run_model(self, model=None, show_plots=False):
        """
        Function used to run DL model. It will train the DL model using callbacks.
        It will retrieve training and validation accuracy mean and  training and validation loss mean.
        It retrieves the performance plots for the model ( training and validation accuracy mean and
        training and validation loss mean) and the summary plot.
        :param model: Keras classifier object.
        :return: model compiled and history object
        """
        start = time.perf_counter()
        if model is None:
            model = self.model
        if self.model is None:
            self.model = model
        self.generate_callbacks()
        self.train_model()

        # todo change this !!!!!
        # history_dict = self.history.history
        # print(history_dict.keys())
        strings = list(self.history.history.keys())
        # ['accuracy', 'loss', 'val_accuracy', 'val_loss']
        # print(strings)
        # metric1 = strings[0]
        # metric2 = strings[2]

        s1 = str(strings[0]) + ' mean: ', np.mean(self.history.history[strings[0]])
        s2 = str(strings[2]) + ' mean: ', np.mean(self.history.history[strings[2]])
        s3 = str(strings[1]) + ' mean: ', np.mean(self.history.history[strings[1]])
        # s3 = 'Training ' + str(metric2) + ' mean: ', np.mean(self.history.history[metric2])
        s4 = str(strings[3]) + ' mean: ', np.mean(self.history.history[strings[3]])

        # s1 = 'Training Accuracy mean: ', np.mean(self.history.history['accuracy'])
        # s1 = 'Training Accuracy mean: ', np.mean(self.history.history['mse'])
        # s2 = 'Validation Accuracy mean: ', np.mean(self.history.history['val_accuracy'])
        # s2 = 'Validation Accuracy mean: ', np.mean(self.history.history['val_mse'])

        print(s1, '\n', s2, '\n', s3, '\n', s4)

        self.model.model.summary()

        final = time.perf_counter()
        run_time = final - start
        # write report
        if self.report_name is not None:
            self._report(['===TRAIN MODELS===\n', self.run_model.__name__])
            self._report([s1, s2, s3, s4])
            l = []
            self.model.model.summary(print_fn=lambda x: l.append(x))
            self._report(l)
            self._report(f"Finished {self.run_model.__name__} in {run_time:.4f} secs\n\n")

        # plot model
        # todo change the report name to not appear the .txt
        if self.report_name:
            path_save_acc = str(self.report_name[:-4] + 'accuracy.png')
            path_save_loss = str(self.report_name[:-4] + 'loss.png')
        else:
            path_save_acc = None
            path_save_loss = None

        plot_summary_accuracy(self.history, path_save=path_save_acc, show=show_plots)  # , metric = metric1
        plot_summary_loss(self.history, path_save=path_save_loss, show=show_plots)
        if self.report_name:
            path_plot = str(self.report_name[:-4] + str('_plot_model.png'))
            plot_model(self.model.model, to_file=path_plot, show_shapes=True, show_layer_names=True)

        else:
            path_plot = None
            plot_model(self.model.model, show_shapes=True, show_layer_names=True)

        return self.model, self.history

    # @timer
    def train_model_cv(self, x_cv, y_cv, cv=5, model=None):
        """
        Function used to cross validate DL models.
        :param x_cv:  data to train model
        :param y_cv: labels of training data
        :param cv: number of folds to cross validation. 5 by default
        :param model: Keras classifier. If None, retrieves model from the class.
        :return: dataframe with scores per fold, and mean and standard deviation of those scores.
        """
        start = time.perf_counter()
        # https://www.kaggle.com/stefanie04736/simple-keras-model-with-k-fold-cross-validation
        if model is not None:
            self.model = model
        # Define per-fold score containers
        scores_complete = []
        accuracy_list = []
        loss_list = []
        folds = StratifiedKFold(n_splits=cv, shuffle=True, random_state=22).split(x_cv, y_cv)
        # print(folds)
        for j, (train_idx, val_idx) in enumerate(folds):
            print('\nFold ', j)
            try:
                self.x_train = x_cv.iloc[train_idx]
                self.x_test = x_cv.iloc[val_idx]
            except:
                self.x_train = x_cv[train_idx]  # todo resolver esta questao. tem q aceitar dataframes listas, numpy....
                self.x_test = x_cv[val_idx]

            self.y_train = y_cv[train_idx]
            self.y_test = y_cv[val_idx]

            self.validation_split = 0.10

            self.generate_callbacks()
            self.train_model()
            scores, report, cm, cm2 = self.score_testset()

            scores_complete.append(scores)
            metrics = scores.keys()
            # # write report
            # if self.report_name is not None:
            #     self._report(['\nFold ', j, '\n'])
            #     # self._report([s1,s2,s3,s4])
            #     # self._report(summary)
            #     self._report('test_score for {} fold'.format(j))
            #     self._report(score, dataframe=True)

        # melhorar esta parte. n chamar o complete evaluate. calcular scores fora. adicionar sensitivity and sensibility
        df = pd.DataFrame(scores_complete)
        df.loc['mean'] = df.mean()
        df.loc['std'] = df.std()

        # https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.cross_validate.html#sklearn.model_selection.cross_validate
        # # evaluate using n-fold cross validation uses the wrapper scikit learn, n consigo testar num dataset a parte
        # kfold = StratifiedKFold(n_splits=cv, shuffle=True, random_state=7)
        # results = cross_val_score(self.model, x_cv, y_cv, cv=kfold, fit_params={'callbacks': self.callbacks})
        # print(results)
        # print(results.mean())
        # print("MLP: %.2f%% (%.2f%%)" % (results.mean() * 100, results.std() * 100))
        #
        # # reportinf(self.history)

        final = time.perf_counter()
        run_time = final - start
        print(f"Finished {self.train_model_cv.__name__} in {run_time:.4f} secs\n\n")
        # write report
        if self.report_name is not None:
            self._report(['===TRAIN MODELS with CV===\n', self.train_model_cv.__name__])
            self._report(df, dataframe=True)
            self._report(df)
            self._report(f"Finished {self.train_model_cv.__name__} in {run_time:.4f} secs\n\n")

        return df

    def get_opt_params(self, param_grid, model, optType='randomizedSearch', cv=10, dataX=None, datay=None,
                       n_iter_search=15, n_jobs=1,
                       scoring=make_scorer(matthews_corrcoef)):
        """
        Function that allows to perform gridSearchCV or randomizedSearchCV in DL models
        :param param_grid: param grid to use in the optimization.
        :param model: keras classifier. model to optimize
        :param optType: optimization type. accept 'randomizedSearch' (default) or 'gridSearch'
        :param cv: number of folds for cross validation. 10 by default.
        :param dataX: data to train the model. If None (default) it will use the train data defined in class
        :param datay: labels to train the model. If None (default) it will use the train labels defined in class
        :param n_iter_search: number of iterations to do when randomizedSearch is performed
        :param n_jobs: n_jobs CPUs to use. 1 by default
        :param scoring: {metrics instance} scoring function built from make_scorer() or a predefined value in string form
            (choose from the scikit-learn`scoring-parameters
            <http://scikit-learn.org/stable/modules/model_evaluation.html#scoring-parameter>`_).
        :return: best model fit and transformed to train data
        """
        start = time.perf_counter()
        print(param_grid)
        if dataX is None:
            dataX = self.x_train
            datay = self.y_train

        po = ParamOptimizer(estimator=model, optType=optType, paramDic=param_grid,
                            dataX=dataX, datay=datay,
                            cv=cv, n_iter_search=n_iter_search, n_jobs=n_jobs, scoring=scoring)
        gs = po.get_opt_params()
        # summaries
        list_to_write_top_3_models = po.report_top_models(gs)
        s1, s2, s3, df = po.report_all_models(gs)  # metrics for the best model. dataframe with all
        # Set the best parameters to the best estimator
        best_classifier = gs.best_estimator_
        self.model = best_classifier
        model_optimized, history_optimized = self.run_model()
        # best_classifier_fit = best_classifier.fit(self.x_train, self.y_train)
        # dnn_simple = best_classifier_fit
        # self.model=dnn_simple
        final = time.perf_counter()
        run_time = final - start

        for line in list_to_write_top_3_models:
            print(line)
        print(s1)
        print('df')
        print(df)
        # write report
        if self.report_name is not None:
            self._report(list_to_write_top_3_models)
            self._report([s1, s2, s3, f"Finished get_opt_params in {run_time:.4f} secs\n\n"])
            self._report(df, dataframe=True, float_format='%.3f')
        return model_optimized

    # todo. change the order of this path, model  (instead of model, path).
    #  because otherwise if someone calls savemodel('...') it will take that as the model and not the path
    def save_model(self, path='model', model=None):
        """
        Function to save model
        :param model: model compiled to save
        :param path: path where to save model. 'model.h5' by default.
        :return:  None
        """
        if model is None:
            model = self.model

        try:
            model.model.save(path)
            print('model saved at {}'.format(path))
        except Exception as e:
            print(str(e))
            print('model not saved')

    def load_model(self, path=''):
        """
        Function to load saved models
        :param path: path where model is saved
        :return:  reconstructed model
        """
        reconstructed_model = tensorflow.keras.models.load_model(path)
        return reconstructed_model

    ####################################################################################################################
    # EVALUATE
    ####################################################################################################################
    # join this functions equal to shalow and deep
    def conf_matrix_seaborn_table(self, conf_matrix=None, classifier=None, path_save='', show=True,
                                  square=True, annot=True, fmt='d', cbar=False, batch_size=None, **params):
        """
        Function to retrieve confusion matrix using seaborn
        :param conf_matrix:
        :param classifier:
        :param path_save:
        :param show:
        :param square:
        :param annot:
        :param fmt:
        :param cbar:
        :param batch_size:
        :param params:
        :return:
        """

        plt.clf()
        if conf_matrix is None:
            y_pred = classifier.predict(self.x_test, batch_size=batch_size)
            mat = confusion_matrix(self.y_test, y_pred)
        else:
            mat = conf_matrix

        sns.heatmap(mat.T, square=square, annot=annot, fmt=fmt, cbar=cbar, **params)
        plt.xlabel('true label')
        plt.ylabel('predicted label')
        if path_save is not None:
            plt.savefig(fname=path_save)
        if show is True:
            plt.show()
        plt.clf()

    def model_simple_evaluate(self, x_test=None, y_test=None, model=None):
        """
        Performs model evaluation with model evaluate function. Returns the loss value & metrics values for the model
        in test mode.
        :param x_test: x dataset where to test. If none will retrieve the dataset from the class.
        :param y_test: labels dataset to test. If none will retrieve the y_test from the class.
        :return: returns the score of the model evaluation list test loss, test acc
        e.g.[0.1309928148984909, 0.963699996471405]
        """
        if x_test is None:
            x_test = self.x_test
            y_test = self.y_test
        if model is None:
            model = self.model
        score = model.model.evaluate(x_test, y_test, batch_size=None, verbose=1,
                                     sample_weight=None, steps=None, callbacks=None,
                                     max_queue_size=10, workers=1, use_multiprocessing=False)
        s1 = ("test loss, test acc:", score)
        print(s1)
        # write report
        if self.report_name is not None:
            self._report(['===simple evaluate===\n', self.model_simple_evaluate.__name__, '\n', s1])

        return score

    # def model_complete_evaluate(self, x_test=None, y_test=None, model=None):
    #     """
    #     Performs complete model evaluation using multiple scoring functions.
    #     :param x_test: x dataset where to test. If none will retrieve the dataset from the class.
    #     :param y_test: labels dataset to test. If none will retrieve the y_test from the class.
    #     :param model: classifier to evaluate. None by default. If None, it will retrieves the trained model from the class
    #     :return: dictionary containing scores, src report, simple confusion matrix, confusion matrix by
    #     class (in binary src this last object will retrieve empty)
    #     """
    #     saved_args = locals()
    #     if x_test is None:
    #         x_test = self.x_test
    #         y_test = self.y_test
    #     if model is None:
    #         model = self.model
    #     scores = {}
    #     cm2 = []
    #
    #     # if batch_size == 1:
    #     #     y_pred = []
    #     #     y_prob = []
    #     #     # online forecast
    #     #     # https://machinelearningmastery.com/use-different-batch-sizes-training-predicting-python-keras/
    #     #     for i in range(x_test.shape[0]):
    #     #         testX, testy = x_test[i], np.array(y_test)[i]
    #     #         y_pre = model.predict(testX, batch_size=1)
    #     #         print(y_pre)
    #     #         y_pred.append(y_pre[0][0])
    #     #         y_pro = model.predict_proba(testX, batch_size=1)
    #     #         y_prob.append(y_pro[0][0])
    #     #     y_pred = list(chain.from_iterable(y_pred))
    #     #     y_prob = list(chain.from_iterable(y_prob))
    #     #
    #     #     print(y_pred)
    #     #     print(y_test)
    #     # else:
    #     y_pred = model.predict(x_test)
    #     try:
    #         y_prob = model.predict_proba(x_test)
    #     except:
    #         y_prob = model.predict(x_test)
    #         y_pred= np.argmax(y_prob,axis=1)
    #
    #     scores['Accuracy'] = accuracy_score(y_test, y_pred)
    #     scores['MCC'] = matthews_corrcoef(y_test, y_pred)
    #     scores['log_loss'] = log_loss(y_test, y_prob)
    #
    #     if self.final_units > 2:
    #         # multiclass
    #         scores['f1 score weighted'] = f1_score(y_test, y_pred, average='weighted')
    #         scores['f1 score macro'] = f1_score(y_test, y_pred, average='macro')
    #         scores['f1 score micro'] = f1_score(y_test, y_pred, average='micro')
    #         scores['roc_auc ovr'] = roc_auc_score(y_test, y_prob, average='weighted', multi_class='ovr')
    #         y_test_reshape = y_test.reshape(y_test.shape[0])  # roc auc ovo was giving error
    #         scores['roc_auc ovo'] = roc_auc_score(y_test_reshape, y_prob, average='weighted', multi_class='ovo')
    #         scores['precision'] = precision_score(y_test, y_pred, average='weighted')
    #         scores['recall'] = recall_score(y_test, y_pred, average='weighted')
    #         cm2 = multilabel_confusion_matrix(y_test, y_pred)
    #
    #     else:
    #         # binary
    #         scores['f1 score'] = f1_score(y_test, y_pred)
    #         scores['roc_auc'] = roc_auc_score(y_test, y_pred)
    #         precision, recall, thresholds = precision_recall_curve(y_test, y_pred)
    #         scores['Precision'] = precision
    #         scores['Recall'] = recall
    #         tn, fp, fn, tp = confusion_matrix(y_test, y_pred).ravel()
    #         # scores['fdr'] = float(fp) / (tp + fp)
    #         scores['sn'] = float(tp) / (tp + fn)
    #         scores['sp'] = float(tn) / (tn + fp)
    #
    #     report = classification_report(y_test, y_pred, output_dict=False)
    #     cm = confusion_matrix(y_test, y_pred)
    #     scores_df = pd.DataFrame(scores.keys(), columns=['metrics'])
    #     scores_df['scores'] = scores.values()
    #
    #     if self.report_name is not None:
    #         self._report(['===SCORING TEST SET ===\n', self.model_complete_evaluate.__name__, saved_args,
    #                       'report\n', report,
    #                       '\n===confusion_matrix===\n', cm,
    #                       '\n===multilabel confusion matrix===\n', cm2,
    #                       '\n===scores report===\n'])
    #
    #         self._report(scores_df, dataframe=True, float_format='%.4f', index=False)
    #         self.conf_matrix_seaborn_table(conf_matrix=cm, path_save=str(self.report_name[:-4] + 'confusion_matrix.png'),
    #                                        show=False)
    #
    #     # print("=== Confusion Matrix ===")
    #     # print(cm)
    #     # print("=== Classification Report ===")
    #     # print(report)
    #     # print('\n')
    #     # print(scores_df)
    #     self.y_test_pred = y_pred
    #     self.y_test_proba = y_prob
    #
    #     return scores, report, cm, cm2
    def score_testset_classification(self, classifier=None):
        """
        Returns the tests set scores for the specified scoring metrics in a ``pandas.DataFrame``. The calculated metrics
        are Matthews correlation coefficient, accuracy, precision, recall, f1 and area under the Receiver-Operator Curve
        (roc_auc). See `sklearn.metrics <http://scikit-learn.org/stable/modules/classes.html#sklearn-metrics-metrics>`_
        for more information.
        :param classifier: {classifier instance} pre-trained classifier used for predictions. If None, will use the
        one trained in the class.
        :return: ``pandas.DataFrame`` containing the cross validation scores for the specified metrics.
        """
        if classifier is not None:  # priority for the classifier in function
            self.model = classifier

        saved_args = locals()
        cm2 = []
        scores = {}
        y_pred = self.model.predict(self.x_test)
        try:
            y_prob = self.model.predict_proba(self.x_test)
        except:
            y_prob = self.model.predict(self.x_test)
            y_pred = np.argmax(y_prob, axis=1)

        # y_pred = classifier.predict(self.x_test)
        # if self.problem_type == 'classification':
        #     y_pred = y_pred.argmax(axis=-1)
        # else:
        #     y_pred = (y_pred > 0.5).astype('int32')
        #
        # try:
        #     y_prob = classifier.predict_proba(self.x_test)
        # except:
        #     y_prob = None
        # y_pred = np.argmax(y_pred, axis=-1)
        scores['Accuracy'] = accuracy_score(self.y_test, y_pred)
        scores['MCC'] = matthews_corrcoef(self.y_test, y_pred)
        if y_prob is not None:
            scores['log_loss'] = log_loss(self.y_test, y_prob)

        if self.final_units > 2:
            # multiclass
            scores['f1 score weighted'] = f1_score(self.y_test, y_pred, average='weighted')
            scores['f1 score macro'] = f1_score(self.y_test, y_pred, average='macro')
            scores['f1 score micro'] = f1_score(self.y_test, y_pred, average='micro')
            if y_prob is not None:
                scores['roc_auc ovr'] = roc_auc_score(self.y_test, y_prob, average='weighted', multi_class='ovr')
                y_test = self.y_test
                # y_test = y_test.reshape(y_test.shape[0])  # roc auc ovo was giving error
                scores['roc_auc ovo'] = roc_auc_score(y_test, y_prob, average='weighted', multi_class='ovo')
            scores['precision'] = precision_score(self.y_test, y_pred, average='weighted')
            scores['recall'] = recall_score(self.y_test, y_pred, average='weighted')
            cm2 = multilabel_confusion_matrix(self.y_test, y_pred)

        else:
            # binary
            scores['f1 score'] = f1_score(self.y_test, y_pred)
            scores['roc_auc'] = roc_auc_score(self.y_test, y_pred)
            precision, recall, thresholds = precision_recall_curve(self.y_test, y_pred)
            scores['Precision'] = precision
            scores['Recall'] = recall
            tn, fp, fn, tp = confusion_matrix(self.y_test, y_pred).ravel()
            scores['fdr'] = float(fp) / (tp + fp)
            scores['sn'] = float(tp) / (tp + fn)
            scores['sp'] = float(tn) / (tn + fp)

        report = classification_report(self.y_test, y_pred, output_dict=False)
        cm = confusion_matrix(self.y_test, y_pred)
        cm2 = None

        if self.report_name is not None:
            self._report(['===SCORING TEST SET ===\n', self.score_testset.__name__, saved_args, 'report\n', report,
                          '\nconfusion_matrix\n', cm, '\nmultilabel confusion matrix\n', cm2, '\nscores report\n'])
            scores_df = pd.DataFrame(scores.keys(), columns=['metrics'])
            scores_df['scores'] = scores.values()
            self._report(scores_df, dataframe=True, float_format='%.4f', index=False)

            self.conf_matrix_seaborn_table(conf_matrix=cm,
                                           path_save=str(self.report_name[-4:] + 'confusion_matrix.png'),
                                           show=False)

        return scores, report, cm, cm2

    # todo add this fucntion?
    # see if it is worth it. put it with not priniting . giving something closer to the src report scikit learn
    def regression_report(self, y_true, y_pred):
        print(y_true)
        print(y_pred)
        # todo acho que nao funciona se forme dadas probabilidades
        try:
            error = y_true - y_pred
            percentil = [5, 25, 50, 75, 95]
            percentil_value = np.percentile(error, percentil)

            metrics = [
                ('mean absolute error', mean_absolute_error(y_true, y_pred)),
                ('median absolute error', median_absolute_error(y_true, y_pred)),
                ('mean squared error', mean_squared_error(y_true, y_pred)),
                ('max error', max_error(y_true, y_pred)),
                ('r2 score', r2_score(y_true, y_pred)),
                ('explained variance score', explained_variance_score(y_true, y_pred))
            ]

            print('Metrics for regression:')
            for metric_name, metric_value in metrics:
                print(f'{metric_name:>25s}: {metric_value: >20.3f}')

            print('\nPercentiles:')
            for p, pv in zip(percentil, percentil_value):
                print(f'{p: 25d}: {pv:>20.3f}')
            return None
        except Exception as e:
            print(e)

    # todo add this function !!!!
    # https://medium.com/analytics-vidhya/mae-mse-rmse-coefficient-of-determination-adjusted-r-squared-which-metric-is-better-cd0326a5697e
    def score_testset_regression(self, classifier=None):
        """
        Returns the tests set scores for the specified scoring metrics in a ``pandas.DataFrame``. The calculated metrics
        are Matthews correlation coefficient, accuracy, precision, recall, f1 and area under the Receiver-Operator Curve
        (roc_auc). See `sklearn.metrics <http://scikit-learn.org/stable/modules/classes.html#sklearn-metrics-metrics>`_
        for more information.
        :param classifier: {classifier instance} pre-trained classifier used for predictions. If None, will use the
        one trained in the class.
        :return: ``pandas.DataFrame`` containing the cross validation scores for the specified metrics.
        """
        if classifier is None:  # priority for the classifier in function
            classifier = self.model
        saved_args = locals()
        cm2 = []
        scores = {}
        y_pred = classifier.predict(self.x_test)
        # todo put abs in negative values? they give negative values so the model try to get the lower score rather than the higher.
        # https://stackoverflow.com/questions/21443865/scikit-learn-cross-validation-negative-values-with-mean-squared-error
        scores['neg_root_mean_squared_error - RMSE'] = abs(mean_squared_error(self.y_test, y_pred, squared=False))
        scores['neg_mean_squared_error - MSE'] = abs(mean_squared_error(self.y_test, y_pred, squared=True))
        scores['neg_mean_absolute_error - MAE'] = abs(mean_absolute_error(self.y_test, y_pred))
        scores['r2'] = r2_score(self.y_test,
                                y_pred)  # It represents the proportion of variance (of y) that has been explained
        # by the independent variables in the model. It provides an indication of goodness of fit and therefore a measure
        # of how well unseen samples are likely to be predicted by the model, through the proportion of explained variance.
        # As such variance is dataset dependent, R² may not be meaningfully comparable across different datasets.
        scores['explained_variance'] = explained_variance_score(self.y_test,
                                                                y_pred)  # The best possible score is 1.0, lower values are worse
        scores['max_error'] = max_error(self.y_test,
                                        y_pred)  # calculates the maximum residual error. aptures the worst case error between the predicted value and the true value

        try:
            scores['neg_mean_squared_log_error - MSLE'] = abs(mean_squared_log_error(self.y_test,
                                                                                     y_pred))  # expected value of the squared logarithmic (quadratic) error or loss.

        except Exception as e:
            print('neg_mean_squared_log_error could not be calculated')
            print(str(e))

        try:
            scores['neg_median_absolute_error'] = abs(median_absolute_error(self.y_test,
                                                                            y_pred))  # robust to outliers. The loss is calculated by taking the median of all absolute differences between the target and the prediction. Do not support multioutput
        except Exception as e:
            print('neg_median_absolute_error could not be calculated')
            print(str(e))

        # todo only for positive values? some of them ? check
        try:
            scores['neg_mean_poisson_deviance'] = abs(mean_poisson_deviance(self.y_test, y_pred))
        except Exception as e:
            print('neg_mean_poisson_deviance could not be calculated')
            print(str(e))
        try:
            scores['neg_mean_gamma_deviance'] = abs(mean_gamma_deviance(self.y_test, y_pred))
        except Exception as e:
            print('neg_mean_gamma_deviance could not be calculated')
            print(str(e))
        try:
            scores['mean_absolute_percentage_error'] = abs(mean_absolute_percentage_error(self.y_test,
                                                                                          y_pred))  # sensitive to relative errors. supports multioutput
        except Exception as e:
            print('mean_absolute_percentage_error could not be calculated')
            print(str(e))

        report = self.regression_report(self.y_test, y_pred)
        cm = None
        cm2 = None

        if self.report_name is not None:
            self._report(['===SCORING TEST SET ===\n', self.score_testset.__name__, saved_args, 'report\n', report,
                          '\nscores report\n'])
            scores_df = pd.DataFrame(scores.keys(), columns=['metrics'])
            scores_df['scores'] = scores.values()
            self._report(scores_df, dataframe=True, float_format='%.4f', index=False)

        return scores, report

    # todo add this. Put text explaining return all the same?

    def score_testset(self, classifier=None):
        if classifier is not None:
            self.model = classifier
        # todo change this
        if self.problem_type != 'regression':
            scores, report, cm, cm2 = self.score_testset_classification(self.model)
            return scores, report, cm, cm2
        elif self.problem_type == 'regression':
            scores, report = self.score_testset_regression(self.model)
            return scores, report

    # https://hackernoon.com/simple-guide-on-how-to-generate-roc-plot-for-keras-classifier-2ecc6c73115a
    # ROC Curves summarize the trade-off between the true positive rate and false positive rate for a predictive model using different probability thresholds.
    # Precision-Recall curves summarize the trade-off between the true positive rate and the positive predictive value for a predictive model using different probability thresholds.
    # ROC curves are appropriate when the observations are balanced between each class, whereas precision-recall curves are appropriate for imbalanced datasets.

    @timer
    def precision_recall_curve(self, show=False, path_save='precision_recall_curve.png', batch_size=None):
        """
        Plot a Precision-Recall curve.
        PRC curves summarize the trade-off between the true positive rate and the positive predictive value for a
        predictive model using different probability thresholds. While ROC curves are appropriate when the observations
        are balanced between each class, PRC are appropriate for imbalanced datasets.
        :param show: whether to show the PRC. false by default
        :param path_save: path to save the plot. if None, will not be stored.
        :param batch_size: batch size to calculate the PRC. None by default
        :return:
        """
        y = self.y_test
        # y_pred = self.model.predict(self.x_test)
        if batch_size == 1:
            y_pred = []
            y_prob = []
            # online forecast
            # https://machinelearningmastery.com/use-different-batch-sizes-training-predicting-python-keras/
            for i in range(self.x_test.shape[0]):
                testX, testy = self.x_test[i], np.array(y)[i]
                y_pro = self.model.predict_proba(testX, batch_size=1)
                y_prob.append(y_pro)
        else:
            y_prob = self.model.predict_proba(self.x_test, batch_size=batch_size)
        # y_prob = self.model.predict_proba(self.x_test, batch_size=batch_size)
        # todo attentio to this ! isnecessary to binarize. and if it is already binarize? should force the input to be binarized or just through an error in the functions that needed to so
        # take this out and receive binarize in the begining
        # checkk the functions iin the models the sparse categorical.
        # Use label_binarize to be multi-label like settings
        # and careful with the binary. take this out of here
        Y = label_binarize(y, classes=np.unique(self.y_train))

        plot_precision_recall_curve(Y, y_prob, self.final_units, show=show, path_save=path_save)

    @timer
    def roc_curve(self, classifier=None, ylim=(0.0, 1.00), xlim=(0.0, 1.0),
                  title='Receiver operating characteristic (ROC) curve',
                  path_save='plot_roc_curve.png', show=False, batch_size=None):
        """
        Plot receiver operating characteristic (ROC) curve.
        ROC Curves summarize the trade-off between the true positive rate and false positive rate for a predictive
        model using different probability thresholds.
        ROC curves are appropriate when the observations are balanced between each class.
        :param classifier: classifier instance. If None, will retrieve the trained one from the class.
        :param ylim: y axis limits
        :param xlim: x axis limits
        :param title: title of the plot. 'Receiver operating characteristic (ROC) curve' by default
        :param path_save: path where to save the plot.  'plot_roc_curve.png' by default
        :param show: Whether to display the produced plot or not. False by default.
        :param batch_size: batch size to use. None by default.
        :return: None
        """
        if classifier is None:
            classifier = self.model

        # attention that do the binarizer inside the function

        plot_roc_curve(classifier, self.final_units, self.x_test, self.y_test, self.x_train, self.y_train, ylim=ylim,
                       xlim=xlim,
                       title=title,
                       path_save=path_save, show=show)

    ###################################################################################################################
    # PREDICT
    ###################################################################################################################

    @timer
    def predict(self, x, seqs=None, classifier=None, names=None, true_y=None, batch=None):
        """
        This function can be used to predict with a trained classifier model. The function returns a
        'pandas.DataFrame' with predictions using the specified estimator and tests data.
        :param x: {array} descriptor values of the peptides to be predicted.
        :param seqs: {list} sequences of the peptides in ``x``.
        :param classifier: {classifier instance} classifier used for predictions. When None ( by default)will use the
        trained one inside class.
        :param names: {list} (optional) names of the peptides in ``x``.
        :param true_y: {array} (optional) true (known) classes of the peptides.
        :param batch: batch size to perform the prediction. None by default.
        :return: ``pandas.DataFrame`` containing predictions for ``x``
       """
        saved_args = locals()
        if classifier is not None:
            self.model = classifier

        predict = self.model.predict(x, batch_size=batch)

        predict_df = pd.DataFrame(columns=['names', 'sequence', 'class predicted', 'True classes'])
        predict_df['names'] = names
        predict_df['True classes'] = true_y
        predict_df['sequence'] = seqs
        predict_df['class predicted'] = predict
        # remove columns with Nans all (cols that were None in entry
        predict_df = predict_df.dropna(axis=1)
        print(predict_df)

        if self.problem_type == 'regression':
            pass
        else:
            try:
                preds = classifier.predict_proba(x, batch_size=batch)
                if self.problem_type == 'binary':
                    cols_prob = ['prob_class_0', 'prob_class_1']
                else:
                    cols_prob = ['prob_class_{}'.format(x + 1) for x in range(self.final_units)]
                preds_df = pd.DataFrame(preds, columns=cols_prob)
                predict_df[cols_prob] = preds
                predict_df = predict_df.round(4)
            except Exception as e:
                print('classifier does not support predict probabilities')
                print(str(e))
                return predict_df

        if self.report_name is not None:
            self._report([self.predict.__name__, saved_args])
            predict_df.to_csv(str(self.report_name + 'predict' + time.strftime("-%Y%m%-%H%M%S.csv")))

        return predict_df

    ####################################################################################################################
    # MODEL CURVES
    # ##################################################################################################################
    # if tensorboard on jupyter is on dont know if it is necessary

    # plot validation curve makes sense?
    @timer
    def plot_validation_curve(self, param_name, param_range,
                              classifier=None,
                              cv=5,
                              score=make_scorer(matthews_corrcoef), title="Validation Curve",
                              xlab="parameter range", ylab="score", n_jobs=1, show=False,
                              path_save='plot_validation_curve', **params):

        """
        This function plots a cross-validation curve for the specified classifier on all tested parameters given in the
        option 'param_range'.

        :param classifier: {classifier instance} classifier or validation curve. If None will use the classifier trianed inside class.
        :param param_name: {string} parameter to assess in the validation curve plot. For example,
        deep learning examples
        :param param_range: {list} parameter range for the validation curve.
        :param cv: {int} number of folds for cross-validation.
        :param score: {metrics instance} scoring function built from make_scorer() or a predefined value in string form
            `sklearn.model_evaluation.scoring-parameter
            <http://scikit-learn.org/stable/modules/model_evaluation.html#scoring-parameter>`_.
        :param title: {str} graphic title. "Validation Curve" by default
        :param xlab: {str} x axis label. "parameter range" by default.
        :param ylab: {str} y axis label. "score" by default
        :param n_jobs: {int} number of parallel jobs to use for calculation. if -1, all available cores are used. 1 by default.
        :param path_save: str. path to save the plot. 'plot_validation_curve' by default.
        :return: plot of the validation curve.
        """
        if classifier is None:
            classifier = self.model

        plot_validation_curve(classifier, self.x_train, self.y_train, param_name, param_range,
                              cv=cv, score=score,
                              title=title, xlab=xlab, ylab=ylab,
                              n_jobs=n_jobs,
                              show=show, path_save=path_save,
                              **params)

    @timer
    def plot_learning_curve(self, classifier=None, title='Learning curve', ylim=None,
                            cv=ShuffleSplit(n_splits=10, test_size=0.2, random_state=0),
                            n_jobs=1, train_sizes=np.linspace(.1, 1.0, 5),
                            path_save='plot_learning_curve', show=False, scalability=True, performance=True):
        """
        Plot a learning curve to determine cross validated training and tests scores for different training set sizes.
        It retrieves graphic representing learning curves, numbers of trainig examples, scores on training sets,
        and scores on tests set and scalability and performance plots if set to True.
        :param classifier: {classifier instance} classifier. If None (default) uses the classifier inside the class.
        :param title: title of the plot
        :param ylim: None by default.
        :param cv: cross validation to use. ShuffleSplit(n_splits=10, test_size=0.2, random_state=0) by default.
        :param n_jobs:  number of parallel jobs to use for calculation. if -1, all available cores are used. default 1.
        :param train_sizes: train sizes to tests
        :param path_save: path to save the plot. If None , lot is not saved. 'plot_learning_curve' by default.
        :param show: Whether to display the graphic. False by default.
        :param scalability: whether to retrieve scalability plots. Fit times per training examples.
        :param performance: whether to retrieve performance plots . Plot fit_time vs score
        :return: None
        https://scikit-learn.org/stable/auto_examples/model_selection/plot_learning_curve.html#sphx-glr-auto-examples-model-selection-plot-learning-curve-py
        """
        if classifier is None:
            classifier = self.model

        plot_learning_curve(classifier, self.x_train, self.y_train, title=title, ylim=ylim,
                            cv=cv,
                            n_jobs=n_jobs, train_sizes=train_sizes,
                            path_save=path_save, show=show, scalability=scalability, performance=performance)

    ####################################################################################################################
    # RUN PREDEFINED MODELS
    # ##################################################################################################################
    # todo passar para outra class?
    def _get_opt_params(self, param_grid, func_name, model, optType, cv, n_iter_search, n_jobs, scoring):
        start = time.perf_counter()
        # retrieve default grids
        if param_grid is None:
            param = param_deep()
            if optType == 'gridSearch':
                param_grid = param[func_name]['param_grid']
            else:
                param_grid = param[func_name]['distribution']

        po = ParamOptimizer(estimator=model, optType=optType, paramDic=param_grid,
                            dataX=self.x_train, datay=self.y_train,
                            cv=cv, n_iter_search=n_iter_search, n_jobs=n_jobs, scoring=scoring)

        gs = po.get_opt_params()
        # summaries
        list_to_write_top_3_models = po.report_top_models(gs)
        s1, s2, s3, df = po.report_all_models(gs)  # metrics for the best model. dataframe with all
        # Set the best parameters to the best estimator
        best_classifier = gs.best_estimator_
        self.model = best_classifier
        model_optimized, history_optimized = self.run_model()
        # best_classifier_fit = best_classifier.fit(self.x_train, self.y_train)
        # dnn_simple = best_classifier_fit
        # self.model=dnn_simple
        final = time.perf_counter()
        run_time = final - start

        print(list_to_write_top_3_models)
        print(s1)
        print('df')
        print(df)
        # write report
        if self.report_name is not None:
            self._report(list_to_write_top_3_models)
            self._report([s1, s2, s3, f"Finished {func_name} in {run_time:.4f} secs\n\n"])
            self._report(df, dataframe=True, float_format='%.3f')
        return model_optimized

    def run_dnn_simple(self,
                       input_dim,
                       optimizer='Adam',
                       hidden_layers=(128, 64),
                       dropout_rate=(0.3,),
                       batchnormalization=(True,),
                       l1=1e-5, l2=1e-4,
                       final_dropout_value=0.3,
                       initial_dropout_value=0.0,
                       loss_fun=None, activation_fun=None,
                       cv=None, optType=None, param_grid=None, n_iter_search=15, n_jobs=1,
                       scoring=make_scorer(matthews_corrcoef)):

        """
        It runs a home network based on stack of Dense layers. It is possible to run a single network, a cross validation
        score or to hyperparameter optimized the network ( with user defined grids or the default ones)
        :param input_dim: input dim for the network
        :param optimizer: optimizer to use when compiling the model. 'Adam' by default.
        :param hidden_layers: tuple. hidden layers to consider. Default (128,64) will run a etwork with 2 hidden layers
        with 128, followed by one with 64.
        :param dropout_rate: tuple. dropout rate to be used after each layer of dense. If dropout rate is only one number,
        it will be propagated to the dense layers existent. By default (0.3,). It will introduces 0.3 dropout rate after
         the first and after the second dense layer.
        :param batchnormalization: tuple. wheteher to use or not batch normalizzation after each dense layer. If
        is only one value it will be propagated to the dense layers existent. By default (True,). It will introduces
        bactchNorm after the first and after the second dense layer.
        :param l1: l1 regularization value
        :param l2: l2 regularization value
        :param final_dropout_value: whether to introduce a final dropout rate before classificaton layer. 0.3 by default.
        :param initial_dropout_value: whether to introduce an initial dropout rate before first layer. 0.0 by default
        :param loss_fun: loss function to consider when compiled by model. If None ( default) it will be retrieved
        accordingly to the problem type (if is binary, multiclass...)
        :param activation_fun: activation function to consider in src layer. If None ( default) it will be retrieved
        accordingly to the problem type (if is binary, multiclass...)
        :param cv: number of folds to consider when doing cross validation or hyperparameter optimization. If None it will
        generate the trainig f a single model.
        :param optType: optimization type to perform. Whether 'gridSearch' or 'randomizedSearch'. If None no optimization is done.
        :param param_grid: param grid to use when optType is not none. If none, param grid by default are retrieved.
        :param n_iter_search: number of iterations in randomizedSearch
        :param n_jobs: 1 by default
        :param scoring: {metrics instance} scoring function built from make_scorer() or a predefined value in string form
            `sklearn.model_evaluation.scoring-paramete
        :return: dnn trained. This dnn trained may be from a single process or the best dnn in the grid space.
        In the case of cross validation, will also retrieve dataframe with scores
        """
        func_name = self.run_dnn_simple.__name__
        if loss_fun is None: loss_fun = self.loss
        if activation_fun is None: activation_fun = self.activation
        saved_args = locals()
        print('dnn model simple')
        model = KerasClassifier(build_fn=create_dnn_simple, number_classes=self.final_units,
                                input_dim=input_dim,
                                optimizer=optimizer,
                                hidden_layers=hidden_layers,
                                dropout_rate=dropout_rate,
                                batchnormalization=batchnormalization,
                                l1=l1, l2=l2,
                                initial_dropout_value=initial_dropout_value,
                                loss_fun=loss_fun, activation_fun=activation_fun,
                                nb_epoch=self.epochs, batch_size=self.batch_size,
                                verbose=self.verbose)

        if cv is not None and optType is None:  # will do cross validation of the model
            if self.report_name is not None:
                self._report(['===Train basic models: \n', self.run_dnn_simple.__name__, saved_args])
            scores = self.train_model_cv(self.x_train, self.y_train, cv=cv, model=model)
            dnn_simple = self.run_model(model)
            return scores

        elif optType is not None:  # will do param_optimization
            # retrieve default grids
            if self.report_name is not None:
                self._report(['===train models with {} param optimization===\n'.format(optType),
                              func_name, saved_args])
            dnn_simple = self._get_opt_params(param_grid, func_name, model, optType, cv, n_iter_search, n_jobs, scoring)

        else:
            if self.report_name is not None:
                self._report(['===Train basic models: \n', func_name, saved_args])
            dnn_simple = self.run_model(model)

        return dnn_simple

    # @timer
    def run_dnn_embedding(self, input_dim,
                          optimizer='Adam',
                          input_dim_emb=21, output_dim=256, input_length=1000, mask_zero=True,
                          hidden_layers=(128, 64),
                          dropout_rate=(0.3,),
                          batchnormalization=(True,),
                          l1=1e-5, l2=1e-4,
                          loss_fun=None, activation_fun=None,
                          cv=None, optType=None, param_grid=None, n_iter_search=15, n_jobs=1,
                          scoring=make_scorer(matthews_corrcoef)):

        func_name = self.run_dnn_embedding.__name__
        if loss_fun is None: loss_fun = self.loss
        if activation_fun is None: activation_fun = self.activation
        saved_args = locals()
        print('dnn model embedding')
        model = KerasClassifier(build_fn=create_dnn_embedding, number_classes=self.final_units,
                                input_dim=input_dim,
                                optimizer=optimizer,
                                input_dim_emb=input_dim_emb, output_dim=output_dim, input_length=input_length,
                                mask_zero=mask_zero,
                                hidden_layers=hidden_layers,
                                dropout_rate=dropout_rate,
                                batchnormalization=batchnormalization,
                                l1=l1, l2=l2,
                                loss_fun=loss_fun, activation_fun=activation_fun,
                                nb_epoch=self.epochs, batch_size=self.batch_size,
                                verbose=self.verbose)

        if cv is not None and optType is None:  # will do cross validation of the model
            dnn_embedding = self.train_model_cv(self.x_train, self.y_train, cv=cv, model=model)
        elif optType is not None:  # will do param_optimization
            if self.report_name is not None:
                self._report(['===train models with {} param optimization===\n'.format(optType),
                              func_name, saved_args])
            dnn_embedding = self._get_opt_params(param_grid, func_name, model, optType, cv, n_iter_search, n_jobs,
                                                 scoring)

        else:
            if self.report_name is not None:
                self._report(['===Train basic models: \n', func_name, saved_args])
            dnn_embedding = self.run_model(model)

        return dnn_embedding

    def run_lstm_simple(self, input_dim,
                        optimizer='Adam',
                        bilstm=True,
                        lstm_layers=(128, 64),
                        dense_layers=(64,),
                        activation='tanh',
                        recurrent_activation='sigmoid',
                        dense_activation="relu",
                        l1=1e-5, l2=1e-4,
                        dropout_rate=(0.3,), recurrent_dropout_rate=(0.3,),
                        dropout_rate_dense=(0.3,),
                        batchnormalization=(True,),
                        loss_fun=None, activation_fun=None,
                        cv=None, optType=None, param_grid=None, n_iter_search=15, n_jobs=1,
                        scoring=make_scorer(matthews_corrcoef)):

        func_name = self.run_lstm_simple.__name__
        if loss_fun is None: loss = self.loss
        if activation_fun is None: activation_fun = self.activation
        saved_args = locals()
        print('lstm1 simple model')

        model = KerasClassifier(build_fn=create_lstm_bilstm_simple,
                                number_classes=self.final_units,
                                input_dim=input_dim,
                                optimizer=optimizer,
                                bilstm=bilstm,
                                lstm_layers=lstm_layers,
                                dense_layers=dense_layers,
                                activation=activation,
                                recurrent_activation=recurrent_activation,
                                dense_activation=dense_activation,
                                dropout_rate=dropout_rate, recurrent_dropout_rate=recurrent_dropout_rate,
                                dropout_rate_dense=dropout_rate_dense,
                                batchnormalization=batchnormalization,
                                l1=l1, l2=l2,
                                loss_fun=loss_fun, activation_fun=activation_fun,
                                nb_epoch=self.epochs, batch_size=self.batch_size, verbose=self.verbose)

        if cv is not None and optType is None:  # will do cross validation of the model
            lstm_simple = self.train_model_cv(self.x_train, self.y_train, cv=cv, model=model)

        elif optType is not None:  # will do param_optimization
            if self.report_name is not None:
                self._report(['===train models with {} param optimization===\n'.format(optType),
                              func_name, saved_args])
            lstm_simple = self._get_opt_params(param_grid, func_name, model, optType, cv, n_iter_search, n_jobs,
                                               scoring)

        else:
            if self.report_name is not None:
                self._report(['===Train basic models: \n', func_name, saved_args])
            lstm_simple = self.run_model(model)

        return lstm_simple

    def run_lstm_embedding(self,
                           optimizer='Adam',
                           input_dim_emb=21, output_dim=128, input_length=1000, mask_zero=True,
                           bilstm=True,
                           lstm_layers=(128, 64),
                           activation='tanh',
                           recurrent_activation='sigmoid',
                           dropout_rate=(0.3,), recurrent_dropout_rate=(0.3,),
                           l1=1e-5, l2=1e-4,
                           dense_layers=(64, 32),
                           dense_activation="relu",
                           dropout_rate_dense=(0.3,),
                           batchnormalization=(True,),
                           loss_fun=None, activation_fun=None,
                           cv=None, optType=None, param_grid=None, n_iter_search=15, n_jobs=1,
                           scoring=make_scorer(matthews_corrcoef)):

        func_name = self.run_lstm_embedding.__name__
        if loss_fun is None: loss_fun = self.loss
        if activation_fun is None: activation_fun = self.activation
        saved_args = locals()
        print('lstm1 embedding model')
        model = KerasClassifier(build_fn=create_lstm_embedding, number_classes=self.final_units,
                                optimizer=optimizer,
                                input_dim_emb=input_dim_emb, output_dim=output_dim, input_length=input_length,
                                mask_zero=mask_zero,
                                bilstm=bilstm,
                                lstm_layers=lstm_layers,
                                dense_layers=dense_layers,
                                activation=activation,
                                recurrent_activation=recurrent_activation,
                                dense_activation=dense_activation,
                                dropout_rate=dropout_rate, recurrent_dropout_rate=recurrent_dropout_rate,
                                l1=l1, l2=l2,
                                dropout_rate_dense=dropout_rate_dense,
                                batchnormalization=batchnormalization,
                                loss_fun=loss_fun, activation_fun=activation_fun,
                                nb_epoch=self.epochs, batch_size=self.batch_size, verbose=self.verbose)

        if cv is not None and optType is None:  # will do cross validation of the model
            lstm_embedding = self.train_model_cv(self.x_train, self.y_train, cv=cv, model=model)

        elif optType is not None:  # will do param_optimization
            if self.report_name is not None:
                self._report(['===train models with {} param optimization===\n'.format(optType),
                              func_name, saved_args])
            lstm_embedding = self._get_opt_params(param_grid, func_name, model, optType, cv, n_iter_search, n_jobs,
                                                  scoring)

        else:
            if self.report_name is not None:
                self._report(['===Train basic models: \n', func_name, saved_args])
            lstm_embedding = self.run_model(model)

        return lstm_embedding

    def run_cnn_lstm(self,
                     input_dim,
                     optimizer='Adam',
                     filter_count=(32, 64, 128),
                     padding='same',
                     strides=1,
                     kernel_size=(3,),
                     cnn_activation=None,
                     kernel_initializer='glorot_uniform',
                     dropout_cnn=(0.0,),
                     max_pooling=(True,),
                     pool_size=(2,), strides_pool=1,
                     data_format_pool='channels_first',
                     bilstm=True,
                     lstm_layers=(128, 64),
                     activation='tanh',
                     recurrent_activation='sigmoid',
                     dropout_rate=(0.3,), recurrent_dropout_rate=(0.3,),
                     l1=1e-5, l2=1e-4,
                     dense_layers=(64, 32),
                     dense_activation="relu",
                     dropout_rate_dense=(0.0,),
                     batchnormalization=(True,),
                     loss_fun=None, activation_fun=None,
                     cv=None, optType=None, param_grid=None, n_iter_search=15, n_jobs=1,
                     scoring=make_scorer(matthews_corrcoef)):

        func_name = self.run_cnn_lstm.__name__
        if loss_fun is None: loss_fun = self.loss
        if activation_fun is None: activation_fun = self.activation
        saved_args = locals()

        print('cnn lstm1 model')
        model = KerasClassifier(build_fn=create_cnn_lstm,
                                number_classes=self.final_units,
                                input_dim=input_dim,
                                optimizer=optimizer,
                                filter_count=filter_count,
                                padding=padding,
                                strides=strides,
                                kernel_size=kernel_size,
                                cnn_activation=cnn_activation,
                                kernel_initializer=kernel_initializer,
                                dropout_cnn=dropout_cnn,
                                max_pooling=max_pooling,
                                pool_size=pool_size, strides_pool=strides_pool,
                                data_format_pool=data_format_pool,
                                bilstm=bilstm,
                                lstm_layers=lstm_layers,
                                dropout_rate=dropout_rate, recurrent_dropout_rate=recurrent_dropout_rate,
                                activation=activation, recurrent_activation=recurrent_activation,
                                dense_layers=dense_layers,
                                dense_activation=dense_activation,
                                l1=l1, l2=l2,
                                dropout_rate_dense=dropout_rate_dense,
                                batchnormalization=batchnormalization,
                                loss_fun=loss_fun, activation_fun=activation_fun,
                                nb_epoch=self.epochs, batch_size=self.batch_size, verbose=self.verbose)

        if cv is not None and optType is None:  # will do cross validation of the model
            cnn_lstm = self.train_model_cv(self.x_train, self.y_train, cv=cv, model=model)

        elif optType is not None:  # will do param_optimization
            if self.report_name is not None:
                self._report(['===train models with {} param optimization===\n'.format(optType),
                              func_name, saved_args])
            cnn_lstm = self._get_opt_params(param_grid, func_name, model, optType, cv, n_iter_search, n_jobs, scoring)

        else:
            if self.report_name is not None:
                self._report(['===Train basic models: \n', func_name, saved_args])

            cnn_lstm = self.run_model(model)
        return cnn_lstm

    @timer
    def run_cnn_1D(self, input_dim,
                   optimizer='Adam',
                   filter_count=(32, 64, 128),  # define number layers
                   padding='same',
                   strides=1,
                   kernel_size=(3,),  # list of kernel sizes per layer. if number will be the same in all numbers
                   cnn_activation='relu',
                   kernel_initializer='glorot_uniform',
                   dropout_cnn=(0.0, 0.2, 0.2),
                   # list of dropout per cnn layer. if number will be the same in all numbers
                   max_pooling=(True,),
                   pool_size=(2,), strides_pool=1,
                   data_format_pool='channels_first',
                   dense_layers=(64, 32),
                   dense_activation="relu",
                   dropout_rate=(0.3,),
                   l1=1e-5, l2=1e-4,
                   loss_fun=None, activation_fun=None,
                   cv=None, optType=None, param_grid=None, n_iter_search=15, n_jobs=1,
                   scoring=make_scorer(matthews_corrcoef)):
        func_name = self.run_cnn_1D.__name__
        if loss_fun is None: loss_fun = self.loss
        if activation_fun is None: activation_fun = self.activation
        saved_args = locals()

        print('cnn 1D model')
        model = KerasClassifier(build_fn=create_cnn_1D, input_dim=input_dim,
                                number_classes=self.final_units,
                                optimizer=optimizer,
                                filter_count=filter_count,  # define number layers
                                padding=padding,
                                strides=strides,
                                kernel_size=kernel_size,
                                # list of kernel sizes per layer. if number will be the same in all numbers
                                cnn_activation=cnn_activation,
                                kernel_initializer=kernel_initializer,
                                dropout_cnn=dropout_cnn,
                                # list of dropout per cnn layer. if number will be the same in all numbers
                                max_pooling=max_pooling,
                                pool_size=pool_size, strides_pool=strides_pool,
                                data_format_pool=data_format_pool,
                                dense_layers=dense_layers,
                                dense_activation=dense_activation,
                                dropout_rate=dropout_rate,
                                l1=l1, l2=l2,
                                loss_fun=loss_fun, activation_fun=activation_fun,
                                nb_epoch=self.epochs, batch_size=self.batch_size, verbose=self.verbose)

        if cv is not None and optType is None:  # will do cross validation of the model
            cnn_1d = self.train_model_cv(self.x_train, self.y_train, cv=cv, model=model)
        elif optType is not None:  # will do param_optimization
            if self.report_name is not None:
                self._report(['===train models with {} param optimization===\n'.format(optType),
                              func_name, saved_args])
            cnn_1d = self._get_opt_params(param_grid, func_name, model, optType, cv, n_iter_search, n_jobs, scoring)

        else:
            if self.report_name is not None:
                self._report(['===Train basic models: \n', func_name, saved_args])

            cnn_1d = self.run_model(model)

        return cnn_1d

    @timer
    def run_cnn_2D(self, input_dim,
                   optimizer='Adam',
                   filter_count=(32, 64, 128),  # define number layers
                   padding='same',
                   strides=1,
                   kernel_size=((3, 3),),  # list of kernel sizes per layer. if number will be the same in all numbers
                   cnn_activation='relu',
                   kernel_initializer='glorot_uniform',
                   dropout_cnn=(0.0, 0.2, 0.2),
                   # list of dropout per cnn layer. if number will be the same in all numbers
                   max_pooling=(True,),
                   pool_size=((2, 2),), strides_pool=1,
                   data_format_pool='channels_first',
                   dense_layers=(64, 32),
                   dense_activation="relu",
                   dropout_rate=(0.3,),
                   l1=1e-5, l2=1e-4,
                   loss_fun=None, activation_fun=None,
                   cv=None, optType=None, param_grid=None, n_iter_search=15, n_jobs=1,
                   scoring=make_scorer(matthews_corrcoef)):
        func_name = self.run_cnn_2D.__name__
        if loss_fun is None: loss_fun = self.loss
        if activation_fun is None: activation_fun = self.activation
        saved_args = locals()

        print('cnn 2D model')
        model = KerasClassifier(build_fn=create_cnn_2D, input_dim=input_dim, number_classes=self.final_units,
                                optimizer=optimizer,
                                filter_count=filter_count,  # define number layers
                                padding=padding,
                                strides=strides,
                                kernel_size=kernel_size,
                                # list of kernel sizes per layer. if number will be the same in all numbers
                                cnn_activation=cnn_activation,
                                kernel_initializer=kernel_initializer,
                                dropout_cnn=dropout_cnn,
                                # list of dropout per cnn layer. if number will be the same in all numbers
                                max_pooling=max_pooling,
                                pool_size=pool_size, strides_pool=strides_pool,
                                data_format_pool=data_format_pool,
                                dense_layers=dense_layers,
                                dense_activation=dense_activation,
                                dropout_rate=dropout_rate,
                                l1=l1, l2=l2, loss_fun=loss_fun, activation_fun=activation_fun,
                                nb_epoch=self.epochs, batch_size=self.batch_size, verbose=self.verbose)

        if cv is not None and optType is None:  # will do cross validation of the model
            cnn_2d = self.train_model_cv(self.x_train, self.y_train, cv=cv, model=model)

        elif optType is not None:  # will do param_optimization
            if self.report_name is not None:
                self._report(['===train models with {} param optimization===\n'.format(optType),
                              func_name, saved_args])
            cnn_2d = self._get_opt_params(param_grid, func_name, model, optType, cv, n_iter_search, n_jobs, scoring)

        else:
            if self.report_name is not None:
                self._report(['===Train basic models: \n', func_name, saved_args])

            cnn_2d = self.run_model(model)

        return cnn_2d

    ####################################################################################################################
    # Hyperparameter optimization
    ####################################################################################################################
    # todo
    # using other tools other than grid/randomized search for both machine learning shallow and deep
    # hyperas , tune
    # https://docs.ray.io/en/master/tune/
    # https://machinelearningmastery.com/hyperopt-for-automated-machine-learning-with-scikit-learn/
    # https://pypi.org/project/hyperas/
    # https://docs.ray.io/en/master/tune/
    # https://keras-team.github.io/keras-tuner/
    # https://github.com/hyperopt/hyperopt
    # https://github.com/optuna/optuna
    # https://scikit-optimize.github.io/stable/
    # https://www.automl.org/book/

    ####################################################################################################################
    # INTERPRETABILITY
    ####################################################################################################################
    # todo
    # important book
    # https://christophm.github.io/interpretable-ml-book/cnn-features.html
    # Alguns packages para interpretabilidade:
    # https://github.com/slundberg/shap
    # https://github.com/kundajelab/deeplift
    # https://github.com/marcotcr/lime
    # https://github.com/sicara/tf-explain
    # https://captum.ai/

    # tensorboard visualization
    # the embeddings using in the google paper

    # shap values
    # https://medium.com/@gabrieltseng/interpreting-complex-models-with-shap-values-1c187db6ec83
    # https://towardsdatascience.com/explain-any-models-with-the-shap-values-use-the-kernelexplainer-79de9464897a
    # https://www.kaggle.com/dansbecker/advanced-uses-of-shap-values
    # @timer
    # def shap(self, model, data):
    #     # Create object that can calculate shap values
    #     explainer = shap.DeepExplainer(model,data)
    #
    #     # Calculate Shap values
    #     shap_values = explainer.shap_values(data)
    #     print(shap_values)
    #     shap.force_plot(explainer.expected_value[1], shap_values[1], data)
    #     shap.summary_plot(shap_values[1], data) # When plotting, we call shap_values[1]. For src problems, there is a separate array of SHAP values for each possible outcome. In this case, we index in to get the SHAP values for the prediction of "True".
    #
    #     shap.dependence_plot('Ball Possession %', shap_values[1], X, interaction_index="Goal Scored")

    # CNN
    # https://towardsdatascience.com/applied-deep-learning-part-4-convolutional-neural-networks-584bc134c1e2
    # Feature maps
    # Convnet filters
    # Class output
    # https://raghakot.github.io/keras-vis/vis.visualization/
    # https://machinelearningmastery.com/how-to-visualize-filters-and-feature-maps-in-convolutional-neural-networks/
    # https://christophm.github.io/interpretable-ml-book/cnn-features.html

####################################################################################################################
# AUTO ML
####################################################################################################################
# todo
# https://www.automl.org/book/


# todo por aqui os pesos do melhor modelo os parametros

# todo por a guardar so o melhor modelo e n de epoch em epoch https://www.tensorflow.org/tutorials/keras/save_and_load


# In the neural network terminology:
# one epoch = one forward pass and one backward pass of all the training examples
# batch size = the number of training examples in one forward/backward pass. The higher the batch size, the more memory space you'll need.
# number of iterations = number of passes, each pass using [batch size] number of examples. To be clear, one pass = one forward pass + one backward pass (we do not count the forward pass and backward pass as two different passes).
# Example: if you have 1000 training examples, and your batch size is 500, then it will take 2 iterations to complete 1 epoch.
#

# https://colah.github.io/posts/2015-08-Understanding-LSTMs/
# https://developers.google.com/machine-learning/crash-course/ml-intro
# https://www.deeplearning.ai/deep-learning-specialization/
# https://www.tensorflow.org/tutorials/keras/save_and_load


#
# if __name__ == '__main__':
#     from src.mlmodels.run_deep_model import prepare_hotencoded
#
#     ecpred_uniref_90 = '/home/amsequeira/deepbio/datasets/ecpred/ecpred_uniprot_uniref_90.csv'
#     alphabet = "ARNDCEQGHILKMFPSTWYV"
#
#     # # prepare data
#     # x_train, x_test, x_dval, y_train, y_test, y_dval, input_dim, fps_x, fps_y = \
#     #     prepare_hotencoded(ecpred_uniref_90, sequence=True, alphabet=alphabet, seq_len=100,
#     #                        parameter='Cross-reference (SUPFAM)', column_label_name='EC number',
#     #                        multiclass=True, multilabel=False,
#     #                        count_negative_as_class=False,
#     #                        model='dnn')
#
#     # # open class
#     # train = ModelTrainEval(x_train, y_train, x_test, y_test, x_dval, y_dval, validation_split=0.25,
#     #                        epochs=500, callbacks=None, reduce_lr=True, early_stopping=True, checkpoint=True,
#     #                        early_stopping_patience=int(30),
#     #                        reduce_lr_patience=50, reduce_lr_factor=0.2, reduce_lr_min=0.00001,
#     #                        path='/home/amsequeira/deepbio/savedModels/weights/',
#     #                        save=False,
#     #                        plot_model_file='/home/amsequeira/deepbio/savedModels/plot_modelmodel_plot.png',
#     #                        verbose=2, batch_size=256)
#
#     # # train dnn
#     # dnn = train.run_dnn_simple(input_dim,
#     #                            optimizer='Adam',
#     #                            hidden_layers=(128, 64),
#     #                            dropout_rate=0.3,
#     #                            batchnormalization=True,
#     #                            l1=1e-5, l2=1e-4,
#     #                            final_dropout_value=0.0,
#     #                            initial_dropout_value=0.0)
#
#     # lstm1 = train.run_lstm_simple(input_dim,
#     #                              optimizer='Adam',
#     #                              bilstm=True,
#     #                              lstm_layers=(128, 64),
#     #                              dense_layers=(64,),
#     #                              activation='tanh',
#     #                              recurrent_activation='sigmoid',
#     #                              dense_activation="relu",
#     #                              l1=1e-5, l2=1e-4,
#     #                              initial_dropout_value=0.0,
#     #                              dropout_rate=0.3, recurrent_dropout_rate=0.3,
#     #                              final_dropout_value=0.0)
#
#     # train.model_simple_evaluate()
#     # scores = train.model_complete_evaluate(plots=True)
#     # print(pd.Series(scores).to_frame('scores'))
#     # train.print_evaluation(x_test, y_test)
#     #
#     # predict = train.model_predict(x_test)
#     # print(predict)
#
#     from src.mlmodels.run_deep_model import get_hot_encoded_sequence, get_final_fps_x_y, \
#         get_parameter_hot_encoded
#     from src.mlmodels.run_deep_model import list_sequences_padded_fps_y
#     from src.mlmodels.utils_run_models import divide_dataset, binarize_labels
#     from src.mlmodels.run_deep_model import normalization, divide_train_test
#
#
#     def prepare_hotencoded_fps(file, sequence=True, alphabet='ARNDCEQGHILKMFPSTWYV', seq_len=1000,
#                                parameter='Cross-reference (Pfam)', column_label_name='ec_number',
#                                multiclass=False, multilabel=False,
#                                count_negative_as_class=True):
#         if sequence:
#             fps_x, fps_y_bin, fps_y, ec_number = get_hot_encoded_sequence(file, alphabet, seq_len)
#         else:
#             pfam_binary, fps_y_bin, fps_y, ec_number = get_parameter_hot_encoded(file, parameter, column_label_name)
#             fps_x = pfam_binary
#
#         # get fps_x and fps_y accordingly of being multiclass or multilabel and number of classes
#         fps_x, fps_y = get_final_fps_x_y(fps_x, fps_y_bin, ec_number, multiclass, multilabel, count_negative_as_class)
#
#         input_dim = fps_x.shape[1]
#         return fps_x, fps_y, input_dim
#
#
#     # prepare data for hot encoded (sequence or pfam domains)
#     fps_x, fps_y, input_dim = prepare_hotencoded_fps(ecpred_uniref_90, sequence=True, alphabet='ARNDCEQGHILKMFPSTWYV',
#                                                      seq_len=100,
#                                                      parameter='Cross-reference (Pfam)', column_label_name='ec_number',
#                                                      multiclass=False, multilabel=False,
#                                                      count_negative_as_class=True)
#
#     # open class
#     train = ModelTrainEval(fps_x, fps_y, validation_split=0.25,
#                            epochs=500, callbacks=None, reduce_lr=True, early_stopping=True, checkpoint=True,
#                            early_stopping_patience=int(30),
#                            reduce_lr_patience=50, reduce_lr_factor=0.2, reduce_lr_min=0.00001,
#                            path='/home/amsequeira/deepbio/savedModels/weights/',
#                            save=False,
#                            plot_model_file='/home/amsequeira/deepbio/savedModels/plot_modelmodel_plot.png',
#                            verbose=2, batch_size=256)
#
#     dnn_cv = train.run_dnn_simple_cv(input_dim,
#                                      optimizer='Adam',
#                                      hidden_layers=(128, 64),
#                                      dropout_rate=0.3,
#                                      batchnormalization=True,
#                                      l1=1e-5, l2=1e-4,
#                                      final_dropout_value=0.3,
#                                      initial_dropout_value=0.2,
#                                      cv=3)



"""
#########################################################################################

A class used for reading sequences.
The main objective is to create dataframes with valid sequences to calculate descriptors.

Authors: João Nuno Abreu 
Date: 03/2022
Email:

#########################################################################################
"""

from .utils import checker, checker_cut
import pandas as pd
import os
from Bio.SeqIO.FastaIO import SimpleFastaParser

class ReadDNA:
    def __init__(self):
        pass

    def read_sequence(self, sequence: str) -> str:
        """
        Reads a sequence, checks if it's valid and returns a dataframe with the sequence.
        """
        if checker(sequence):
            return sequence.strip().upper()
        else:
            raise ValueError("Error! Invalid character in sequence:", sequence)

    def read_fasta(self, filename: str, with_labels: bool = False) -> pd.DataFrame:
        """
        Reads the input file in fasta format. 
        It always reads sequences, and labels if the user wants.
        If the user wants the labels, he must specify the with_labels parameter as True and the FASTA format must be the following:
            >sequence_id1,label1
            ACTGACTG...
            >sequence_id2,label2
            ACTGACTG...
        """
        labels = []
        sequences = []
        
        if not os.path.isfile(filename):
            raise ValueError("Error! File does not exist:", filename)

        if 'fasta' not in filename:
            raise ValueError("Error! File must be in fasta format:", filename)
        
        with open(filename) as handle:
            for key, sequence in SimpleFastaParser(handle):
                # get label and check if it's valid
                if with_labels:
                    label = int(key.split(',')[1])
                    if(label not in [0,1]):
                        raise ValueError("Error! Label must be either 0 or 1 and it is:", label)
                    else:
                        labels.append(label)
                
                # get sequence and check if it's valid
                sequence = sequence.strip().upper()
                if checker(sequence):
                    sequences.append(sequence)      
                else:
                    raise ValueError("Error! Invalid character in sequence:", key)
        
        # add labels to result if the user wants
        if with_labels:
            return pd.DataFrame(list(zip(sequences, labels)), columns=['sequence', 'label'])
        else:
            return pd.DataFrame(sequences, columns=['sequence'])
            
    
    def read_csv(self, filename: str, with_labels: bool = False) -> pd.DataFrame:
        """
        Reads the input file in csv format. 
        It always reads sequences, and labels if the user wants. 
        There must be a column with the sequence.
        If the user wants the labels, he must specify the with_labels parameter as True and the column with the labels must be named "label".
        """
        
        if not os.path.isfile(filename):
            raise ValueError("Error! File does not exist:", filename)

        if 'csv' not in filename:
            raise ValueError("Error! File must be in csv format:", filename)
        
        dataset = pd.read_csv(filename)
        
        # check column names
        if 'sequence' not in dataset.columns:
            raise ValueError("The dataset must always have the column 'sequence'")
        if with_labels and 'label' not in dataset.columns:
            raise ValueError("Since with_labels is True, the dataset must have the column 'label'")
        
        # get sequences and labels
        sequences = dataset['sequence'].to_list()
        
        if with_labels:
            labels = dataset['label'].to_list()
        
        # check if sequences are valid
        valid_sequences = []
        for sequence in sequences:
            if checker(sequence) and "cut" not in filename:
                valid_sequences.append(sequence.strip().upper())  
            elif checker_cut(sequence) and "cut" in filename:
                valid_sequences.append(sequence.strip().upper())  
            else:
                raise ValueError("Error! Invalid character in sequence:", sequence)
            
        # check if labels are valid
        valid_labels = []
        if with_labels:
            for label in labels:
                if(label not in [0,1]):
                    raise ValueError("Error! Label must be either 0 or 1 and it is:", label)
                else:
                    valid_labels.append(label)
        
        # add labels to result if the user wants
        if with_labels:
            return pd.DataFrame(list(zip(sequences, labels)), columns=['sequence', 'label'])
        else:
            return pd.DataFrame(sequences, columns=['sequence'])
            
        
if __name__ == "__main__":
    reader = ReadDNA()
    data = reader.read_csv('datasets/primer/dataset.csv', with_labels=True)
    print(data)
    
    data = reader.read_fasta('datasets/primer/dataset.fasta', with_labels=True)
    print(data)
    
    


# -*- coding: utf-8 -*-
#!/usr/bin/env python
"""
##############################################################################
This function gives binary profile of residues for all the  sequences for 25 phychem feature

It receives a sequence and a list containing feature numbers (valid number, 0-24)

    \n FEATURE NAME                    FEATURE NUMBER

    \n 'Positively charged' --                  0
    \n 'Negatively charged' --                  1
    \n 'Neutral charged' --                     2
    \n 'Polarity' --                            3
    \n 'Non polarity' --                        4
    \n 'Aliphaticity' --                        5
    \n 'Cyclic' --                              6
    \n 'Aromaticity' --                         7
    \n 'Acidicity'--                            8
    \n 'Basicity'--                             9
    \n 'Neutral (ph)' --                       10
    \n 'Hydrophobicity' --                     11
    \n 'Hydrophilicity' --                     12
    \n 'Neutral' --                            13
    \n 'Hydroxylic' --                         14
    \n 'Sulphur content' -                     15
    \n 'Secondary Structure(Helix)'            16
    \n 'Secondary Structure(Strands)',         17
    \n 'Secondary Structure(Coil)',            18
    \n 'Solvent Accessibilty (Buried)',        19
    \n 'Solvent Accesibilty(Exposed)',         20
    \n 'Solvent Accesibilty(Intermediate)',    21
    \n 'Tiny',                                 22
    \n 'Small',                                23
    \n 'Large'                                 24

The code is based on the package Pfeature:
Pande, Akshara & Patiyal, Sumeet & Lathwal, Anjali & Arora, Chakit & Kaur, Dilraj & Dhall, Anjali & Mishra, Gaurav & Kaur,
Harpreet & Sharma, Neelam & Jain, Shipra & Usmani, Salman & Agrawal, Piyush & Kumar, Rajesh & Kumar, Vinod & Raghava, Gajendra.
(2019). Computing wide range of protein/peptide features from their sequence and structure. 10.1101/599126.


It returns a dictionary form with the values

Authors: Ana Marta Sequeira

Date: 05/2019

Email:

##############################################################################
"""

import pandas as pd
import numpy as np
import os
# import sys


def init(path=None, index=''):
    """
    Read in files. You need to run this (once) before you can
    access any records. If the files are not within the current directory,
    you need to specify the correct directory path.

    :param path:
    :param index:
    :return:
    """
    index = str(index)
    if path is None:
        for path in [os.path.split(__file__)[0]]:
            if os.path.exists(os.path.join(path, index)):
                break
        # print('path =', path, file=sys.stderr)

    doc=''
    if index == 'bonds.csv':
        doc=(path + '/data/bonds.scv')
    if index == 'PhysicoChemical.csv':
        doc=(path + '/data/PhysicoChemical.csv')
    return doc


doc=init(index='PhysicoChemical.csv')
# Finding physico-chemical property of a vector of polypeptides
PCP = pd.read_csv(doc, header=None)

#headerCT = ['Positively charged_CT_1', 'Positively charged_CT_2', 'Positively charged_CT_3', 'Positively charged_CT_4', 'Positively charged_CT_5', 'Positively charged_CT_6', 'Positively charged_CT_7', 'Positively charged_CT_8', 'Positively charged_CT_9', 'Positively charged_CT_10', 'Negatively charged_CT_1', 'Negatively charged_CT_2', 'Negatively charged_CT_3', 'Negatively charged_CT_4', 'Negatively charged_CT_5', 'Negatively charged_CT_6', 'Negatively charged_CT_7', 'Negatively charged_CT_8', 'Negatively charged_CT_9', 'Negatively charged_CT_10', 'Neutral charged_CT_1', 'Neutral charged_CT_2', 'Neutral charged_CT_3', 'Neutral charged_CT_4', 'Neutral charged_CT_5', 'Neutral charged_CT_6', 'Neutral charged_CT_7', 'Neutral charged_CT_8', 'Neutral charged_CT_9', 'Neutral charged_CT_10', 'Polarity_CT_1', 'Polarity_CT_2', 'Polarity_CT_3', 'Polarity_CT_4', 'Polarity_CT_5', 'Polarity_CT_6', 'Polarity_CT_7', 'Polarity_CT_8', 'Polarity_CT_9', 'Polarity_CT_10', 'Non polarity_CT_1', 'Non polarity_CT_2', 'Non polarity_CT_3', 'Non polarity_CT_4', 'Non polarity_CT_5', 'Non polarity_CT_6', 'Non polarity_CT_7', 'Non polarity_CT_8', 'Non polarity_CT_9', 'Non polarity_CT_10', 'Aliphaticity_CT_1', 'Aliphaticity_CT_2', 'Aliphaticity_CT_3', 'Aliphaticity_CT_4', 'Aliphaticity_CT_5', 'Aliphaticity_CT_6', 'Aliphaticity_CT_7', 'Aliphaticity_CT_8', 'Aliphaticity_CT_9', 'Aliphaticity_CT_10', 'Cyclic_CT_1', 'Cyclic_CT_2', 'Cyclic_CT_3', 'Cyclic_CT_4', 'Cyclic_CT_5', 'Cyclic_CT_6', 'Cyclic_CT_7', 'Cyclic_CT_8', 'Cyclic_CT_9', 'Cyclic_CT_10', 'Aromaticity_CT_1', 'Aromaticity_CT_2', 'Aromaticity_CT_3', 'Aromaticity_CT_4', 'Aromaticity_CT_5', 'Aromaticity_CT_6', 'Aromaticity_CT_7', 'Aromaticity_CT_8', 'Aromaticity_CT_9', 'Aromaticity_CT_10', 'Acidicity_CT_1', 'Acidicity_CT_2', 'Acidicity_CT_3', 'Acidicity_CT_4', 'Acidicity_CT_5', 'Acidicity_CT_6', 'Acidicity_CT_7', 'Acidicity_CT_8', 'Acidicity_CT_9', 'Acidicity_CT_10', 'Basicity_CT_1', 'Basicity_CT_2', 'Basicity_CT_3', 'Basicity_CT_4', 'Basicity_CT_5', 'Basicity_CT_6', 'Basicity_CT_7', 'Basicity_CT_8', 'Basicity_CT_9', 'Basicity_CT_10', 'Neutral (ph)_CT_1', 'Neutral (ph)_CT_2', 'Neutral (ph)_CT_3', 'Neutral (ph)_CT_4', 'Neutral (ph)_CT_5', 'Neutral (ph)_CT_6', 'Neutral (ph)_CT_7', 'Neutral (ph)_CT_8', 'Neutral (ph)_CT_9', 'Neutral (ph)_CT_10', 'Hydrophobicity_CT_1', 'Hydrophobicity_CT_2', 'Hydrophobicity_CT_3', 'Hydrophobicity_CT_4', 'Hydrophobicity_CT_5', 'Hydrophobicity_CT_6', 'Hydrophobicity_CT_7', 'Hydrophobicity_CT_8', 'Hydrophobicity_CT_9', 'Hydrophobicity_CT_10', 'Hydrophilicity_CT_1', 'Hydrophilicity_CT_2', 'Hydrophilicity_CT_3', 'Hydrophilicity_CT_4', 'Hydrophilicity_CT_5', 'Hydrophilicity_CT_6', 'Hydrophilicity_CT_7', 'Hydrophilicity_CT_8', 'Hydrophilicity_CT_9', 'Hydrophilicity_CT_10', 'Neutral_CT_1', 'Neutral_CT_2', 'Neutral_CT_3', 'Neutral_CT_4', 'Neutral_CT_5', 'Neutral_CT_6', 'Neutral_CT_7', 'Neutral_CT_8', 'Neutral_CT_9', 'Neutral_CT_10', 'Hydroxylic_CT_1', 'Hydroxylic_CT_2', 'Hydroxylic_CT_3', 'Hydroxylic_CT_4', 'Hydroxylic_CT_5', 'Hydroxylic_CT_6', 'Hydroxylic_CT_7', 'Hydroxylic_CT_8', 'Hydroxylic_CT_9', 'Hydroxylic_CT_10', 'Sulphur content_CT_1', 'Sulphur content_CT_2', 'Sulphur content_CT_3', 'Sulphur content_CT_4', 'Sulphur content_CT_5', 'Sulphur content_CT_6', 'Sulphur content_CT_7', 'Sulphur content_CT_8', 'Sulphur content_CT_9', 'Sulphur content_CT_10', 'Secondary Structure(Helix)_CT_1', 'Secondary Structure(Helix)_CT_2', 'Secondary Structure(Helix)_CT_3', 'Secondary Structure(Helix)_CT_4', 'Secondary Structure(Helix)_CT_5', 'Secondary Structure(Helix)_CT_6', 'Secondary Structure(Helix)_CT_7', 'Secondary Structure(Helix)_CT_8', 'Secondary Structure(Helix)_CT_9', 'Secondary Structure(Helix)_CT_10', 'Secondary Structure(Strands)_CT_1', 'Secondary Structure(Strands)_CT_2', 'Secondary Structure(Strands)_CT_3', 'Secondary Structure(Strands)_CT_4', 'Secondary Structure(Strands)_CT_5', 'Secondary Structure(Strands)_CT_6', 'Secondary Structure(Strands)_CT_7', 'Secondary Structure(Strands)_CT_8', 'Secondary Structure(Strands)_CT_9', 'Secondary Structure(Strands)_CT_10', 'Secondary Structure(Coil)_CT_1', 'Secondary Structure(Coil)_CT_2', 'Secondary Structure(Coil)_CT_3', 'Secondary Structure(Coil)_CT_4', 'Secondary Structure(Coil)_CT_5', 'Secondary Structure(Coil)_CT_6', 'Secondary Structure(Coil)_CT_7', 'Secondary Structure(Coil)_CT_8', 'Secondary Structure(Coil)_CT_9', 'Secondary Structure(Coil)_CT_10', 'Solvent Accessibilty (Buried)_CT_1', 'Solvent Accessibilty (Buried)_CT_2', 'Solvent Accessibilty (Buried)_CT_3', 'Solvent Accessibilty (Buried)_CT_4', 'Solvent Accessibilty (Buried)_CT_5', 'Solvent Accessibilty (Buried)_CT_6', 'Solvent Accessibilty (Buried)_CT_7', 'Solvent Accessibilty (Buried)_CT_8', 'Solvent Accessibilty (Buried)_CT_9', 'Solvent Accessibilty (Buried)_CT_10', 'Solvent Accesibilty(Exposed)_CT_1', 'Solvent Accesibilty(Exposed)_CT_2', 'Solvent Accesibilty(Exposed)_CT_3', 'Solvent Accesibilty(Exposed)_CT_4', 'Solvent Accesibilty(Exposed)_CT_5', 'Solvent Accesibilty(Exposed)_CT_6', 'Solvent Accesibilty(Exposed)_CT_7', 'Solvent Accesibilty(Exposed)_CT_8', 'Solvent Accesibilty(Exposed)_CT_9', 'Solvent Accesibilty(Exposed)_CT_10', 'Solvent Accesibilty(Intermediate)_CT_1', 'Solvent Accesibilty(Intermediate)_CT_2', 'Solvent Accesibilty(Intermediate)_CT_3', 'Solvent Accesibilty(Intermediate)_CT_4', 'Solvent Accesibilty(Intermediate)_CT_5', 'Solvent Accesibilty(Intermediate)_CT_6', 'Solvent Accesibilty(Intermediate)_CT_7', 'Solvent Accesibilty(Intermediate)_CT_8', 'Solvent Accesibilty(Intermediate)_CT_9', 'Solvent Accesibilty(Intermediate)_CT_10', 'Tiny_CT_1', 'Tiny_CT_2', 'Tiny_CT_3', 'Tiny_CT_4', 'Tiny_CT_5', 'Tiny_CT_6', 'Tiny_CT_7', 'Tiny_CT_8', 'Tiny_CT_9', 'Tiny_CT_10', 'Small_CT_1', 'Small_CT_2', 'Small_CT_3', 'Small_CT_4', 'Small_CT_5', 'Small_CT_6', 'Small_CT_7', 'Small_CT_8', 'Small_CT_9', 'Small_CT_10', 'Large_CT_1', 'Large_CT_2', 'Large_CT_3', 'Large_CT_4','Large_CT_5', 'Large_CT_6', 'Large_CT_7', 'Large_CT_8', 'Large_CT_9', 'Large_CT_10'];
headers = ['Positively charged',
           'Negatively charged',
           'Neutral charged',
           'Polarity',
           'Non polarity',
           'Aliphaticity',
           'Cyclic',
           'Aromaticity',
           'Acidicity',
           'Basicity',
           'Neutral (ph)',
           'Hydrophobicity',
           'Hydrophilicity',
           'Neutral',
           'Hydroxylic',
           'Sulphur content',
           'Secondary Structure(Helix)',
           'Secondary Structure(Strands)',
           'Secondary Structure(Coil)',
           'Solvent Accessibilty (Buried)',
           'Solvent Accesibilty(Exposed)',
           'Solvent Accesibilty(Intermediate)',
           'Tiny',
           'Small',
           'Large']


def encode(peptide):
    """
    encode AA in sequence

    :param peptide: protein sequence
    :return: protein sequence with encoded AA
    """
    #letter = {'A','C','D','E','F','G','H','I','K','L','M','N','P','Q','R','S','T','V','W','Y'};
    l=len(peptide)
    encoded=np.zeros(l)
    for i in range(l):
        if(peptide[i]=='A'):
            encoded[i] = 0
        elif(peptide[i]=='C'):
            encoded[i] = 1
        elif(peptide[i]=='D'):
            encoded[i] = 2
        elif(peptide[i]=='E'):
            encoded[i] = 3
        elif(peptide[i]=='F'):
            encoded[i] = 4
        elif(peptide[i]=='G'):
            encoded[i] = 5
        elif(peptide[i]=='H'):
            encoded[i] = 6
        elif(peptide[i]=='I'):
            encoded[i] = 7
        elif(peptide[i]=='K'):
            encoded[i] = 8
        elif(peptide[i]=='L'):
            encoded[i] = 9
        elif(peptide[i]=='M'):
            encoded[i] = 10
        elif(peptide[i]=='N'):
            encoded[i] = 11
        elif(peptide[i]=='P'):
            encoded[i] = 12
        elif(peptide[i]=='Q'):
            encoded[i] = 13
        elif(peptide[i]=='R'):
            encoded[i] = 14
        elif(peptide[i]=='S'):
            encoded[i] = 15
        elif(peptide[i]=='T'):
            encoded[i] = 16
        elif(peptide[i]=='V'):
            encoded[i] = 17
        elif(peptide[i]=='W'):
            encoded[i] = 18
        elif(peptide[i]=='Y'):
            encoded[i] = 19
        else:
            print('Wrong residue!')
    return encoded


def lookup(peptide, feature_num):
    """

    :param peptide: protein sequence
    :param feature_num: number of the feature to be calculated
    :return: value of property for each AA in sequence
    """
    l=len(peptide)
    peptide = list(peptide)
    out=[]
    peptide_num = encode(peptide)

    for i in range(l):
        #out[i] = PCP[peptide_num[i]][featureNum]
        out.append(PCP[peptide_num[i]][feature_num])
    return out


def binary_profile_all(seq, feature_numb):
    """
    binary profile of physicalchemical characterisitics choosed of a sequence

    :param seq: protein sequence
    :param feature_numb: number of the feature to be calculated
    :return: value of property for each AA in sequence
    """
    l = len(seq)
    #print('length is',l)
    seq=[seq[i].upper() for i in range(l)]
    bin_prof = []
    for i in range(0,l):
        temp = lookup(seq[i], feature_numb)
        bin_prof.append(temp)
    out = pd.DataFrame(bin_prof)

    return bin_prof


def bin_pc_wp(sequence, feature_numb):
    """
    Binary profile of a sequence based on physicalchemical characterisitics choosed

    :param sequence: protein sequence
    :param feature_numb: list with numbers of the features to be calculated
    :return: dictionary with binary profile of a sequence for a list of desired features
    """
    res={}
    for feature in feature_numb:
        binprof = binary_profile_all(sequence,feature)
        for j in range(len(binprof)):
            name_feature='bin_proper_'+str(feature)+'_'+str(j)
            res[name_feature]=binprof[j][0]

    return res


if __name__=="__main__":
    print(bin_pc_wp('AYARY',[24]))
    print(bin_pc_wp('AAACPPPQSTY',[24,5,2,0]))
    print(bin_pc_wp('APACMPPQSTY',[24,5,2,0]))





import numpy as np
import os

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
# from silence_tensorflow import silence_tensorflow
# silence_tensorflow()
# import logging
# logging.disable(logging.WARNING)
# logging.getLogger("tensorflow").setLevel(logging.FATAL)
import tensorflow as tf

from tensorflow.keras.models import Sequential, load_model
from tensorflow.keras.layers import Dense, Dropout, LSTM, Embedding, TimeDistributed, Bidirectional
from tensorflow.keras.wrappers.scikit_learn import KerasClassifier
from tensorflow.keras import regularizers, Input
from tensorflow.keras.layers import BatchNormalization, Flatten
from tensorflow.keras.callbacks import ReduceLROnPlateau, EarlyStopping, ModelCheckpoint
from sklearn.model_selection import StratifiedKFold
from sklearn.model_selection import cross_val_score
from tensorflow.keras.preprocessing.sequence import pad_sequences

# from src.mlmodels.utils import timer, summary_loss, saveModel, loadDNN, reportinf, validation, validation_multiclass
from tensorflow.keras.utils import plot_model
from tensorflow import keras
from operator import itemgetter

tf.debugging.set_log_device_placement(True)
strategy = tf.distribute.MirroredStrategy()


############################################################################################
############################################################################################


def create_dnn_simple(input_dim, number_classes,
                      hidden_layers=(128, 64, 32),
                      optimizer='Adam',
                      initial_dropout_value=0.0,
                      dropout_rate=0.3,
                      batchnormalization=True, activation="relu",
                      l1=1e-5, l2=1e-4,
                      final_dropout_value=0.2):
    last_layer = hidden_layers[-1]
    with strategy.scope():
        model = Sequential()
        model.add(Input(shape=(input_dim,)))



        # https://towardsdatascience.com/applied-deep-learning-part-3-autoencoders-1c083af4d798


"""
##############################################################################

A class  used for computing different types of protein encoddings parallelized.
It contains encodings such nfl, blossum, zcale, one-hot-encoding. It also allows to perform padding.

Authors: Miguel Barros

Date: 03/2022

Email:

##############################################################################
"""
import os

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from tensorflow.keras.preprocessing.sequence import pad_sequences
from tensorflow.keras.utils import to_categorical
import torch
from transformers import AutoTokenizer, AutoModel


class Encoding:
    def __init__(self, dataset, col: str = 'sequence'):
        """
        Constructor

        :param dataset: the data corresponding to the protein sequences, it should be an string (one sequence),
         a list  or a pandas dataframe (multiples sequences).
        :param col: the name of the column in the dataframe which contains the protein sequences (pandas dataframe),
        or the name to give to the protein sequence column (list or string). Default collumn name is 'sequence'.
        """
        if isinstance(dataset, pd.DataFrame):
            self.result = dataset
        elif isinstance(dataset, str):
            data = {col: [dataset]}
            self.result = pd.DataFrame(data)
        elif isinstance(dataset, list):
            data = {col: dataset}
            self.result = pd.DataFrame(data)
        else:
            raise Exception('Parameter dataframe is not an string, list or pandas Dataframe')

        self.col = col
        self.padded = False
        self.result.drop_duplicates(subset=self.col, keep='first', inplace=True)

        for path in [os.path.split(__file__)[0]]:
            if os.path.exists(os.path.join(path, str(''))):
                self.path = path
                break

    def get_seq_pad(self, seq_len: int = 600, alphabet: str = "XARNDCEQGHILKMFPSTWYV", padding_truncating='post',
                    n_jobs: int = 4):
        """
       This methods performs the padding of the sequences.

       :param seq_len: The maximum length for all sequences. By default the length is 600 amino acids
       :param alphabet: The alphabet of aminoacids to be used.
       :param padding_truncating: 'pre' or 'post' pad either before or after each sequence, also removes values from sequences larger than seq_len. By default padding is after each sequence.
       :param n_jobs: number of CPU cores to be used. Default used is 4 CPU cores

       :return: Dataframe with the padded sequences for each sequence.
       """

        if alphabet[0] != 'X':
            alphabet = alphabet.replace('X', '')
            alphabet = 'X' + alphabet
        char_to_int = dict((c, i) for i, c in enumerate(alphabet))  # dicionarios de hot encoding
        int_to_char = dict((i, c) for i, c in enumerate(alphabet))  # dicionarios de reversão
        try:
            with Parallel(n_jobs=n_jobs) as parallel:
                res = parallel(
                    delayed(seq_padding)(seq, char_to_int, int_to_char, seq_len, padding_truncating) for seq
                    in self.result[self.col])
            self.result = self.result.assign(padded_sequence=res)
            return self.result
        except:
            raise Exception('Sequences not preprocessed. Run sequence preprocessing')


    def get_hot_encoded(self, alphabet: str = "XARNDCEQGHILKMFPSTWYV", n_jobs: int = 4):
        """
        This methods performs the one-hot-encoding of sequences.

        :param alphabet: The alphabet of aminoacids to be used.
        :param n_jobs: number of CPU cores to be used. Default used is 4 CPU cores

        :return: Dataframe with the one-hot-encoding for each sequence
        """

        if alphabet[0] != 'X':
            alphabet = alphabet.replace('X', '')
            alphabet = 'X' + alphabet
        char_to_int = dict((c, i) for i, c in enumerate(alphabet))  # dicionario de hot encoding
        try:
            with Parallel(n_jobs=n_jobs) as parallel:
                res = parallel(delayed(seq_hot_encoded)(seq, char_to_int) for seq in self.result[self.col])

            self.result = self.result.assign(One_hot_encoding=res)
            return self.result
        except:
            raise Exception('Amino acids not included in the alphabet found. Please run preprocessing methods.')

    def get_pad_and_hot_encoding(self, seq_len: int = 600, alphabet: str = "XARNDCEQGHILKMFPSTWYV",
                                 padding_truncating='post', n_jobs: int = 4):
        """
        This methods performs the padding of the sequences and one-hot-encode the padded sequences.

        :param seq_len: The maximum length for all sequences. By default the length is 600 amino acids
        :param alphabet: The alphabet of aminoacids to be used.
        :param padding_truncating: 'pre' or 'post' pad either before or after each sequence, also removes values from sequences larger than seq_len. By default padding is after each sequence.
        :param n_jobs: number of CPU cores to be used. Default used is 4 CPU cores

        :return: Dataframe with the padded sequences and the one-hot-encoding for each sequence
        """
        if alphabet[0] != 'X':
            alphabet = alphabet.replace('X', '')
            alphabet = 'X' + alphabet
        char_to_int = dict((c, i) for i, c in enumerate(alphabet))  # dicionarios de hot encoding
        int_to_char = dict((i, c) for i, c in enumerate(alphabet))  # dicionarios de reversão
        # try:
        with Parallel(n_jobs=n_jobs) as parallel:
            res = parallel(delayed(seq_padded_hot)(seq, char_to_int, int_to_char, seq_len,
                                                   padding_truncating) for seq in self.result[self.col])
        pad_seques, hot_enc_seqs = [], []
        for elemt in res:
            pad_seques.append(elemt[1])
            hot_enc_seqs.append(elemt[0])

        self.result = self.result.assign(pad_seques=pad_seques)
        self.result = self.result.assign(One_hot_encoding=hot_enc_seqs)
        return self.result

        # except:
        #     raise Exception('Amino acids not included in the alphabet found. Please run preprocessing methods.')

    def get_blosum(self, blosum: str = 'blosum62', n_jobs: int = 4):
        """
        BLOSUM62 is a substitution matrix that specifies the similarity of one amino acid to another by means of a score.
        This score reflects the frequency of substitutions found from studying protein sequence conservation
        in large databases of related proteins.
        The number 62 refers to the percentage identity at which sequences are clustered in the analysis.
        I is possible to get blosum50 to get 50 % identity.
        Encoding a peptide this way means we provide the column from the blosum matrix corresponding to the amino acid
        at each position of the sequence. This produces 24 * seqlen matrix.

        :param blosum: blosum matrix to use either 'blosum62' or 'blosum50'. by default 'blosum62'
        :param n_jobs: number of CPU cores to be used. Default used is 4 CPU cores

        :return: Dataframe with blosum encoding for each sequence
        """
        if blosum == 'blosum62':
            encoding = pd.read_csv(self.path + '/adjuv_functions/features_functions/data/blosum62.csv',
                                   index_col=0).to_dict()
        elif blosum == 'blosum50':
            encoding = pd.read_csv(self.path + '/adjuv_functions/features_functions/data/blosum50.csv',
                                   index_col=0).to_dict()
        else:
            raise Exception('The provided encoding is not valid')
        try:
            if isinstance(encoding, dict):
                with Parallel(n_jobs=n_jobs) as parallel:
                    res = parallel(delayed(seq_blosum_encoding)(seq, encoding) for seq in self.result[self.col])
                self.result = self.result.assign(blosum=res)
                return self.result
            else:
                raise Exception('The provided encoding is not a dictionary')
        except:
            raise Exception('Sequences not preprocessed. Run sequence preprocessing')

    def get_nlf(self, n_jobs: int = 4):
        """
        Method that takes many physicochemical properties and transforms them using a Fisher Transform (similar to a PCA)
        creating a smaller set of features that can describe the amino acid just as well.
        There are 19 transformed features.
        This method of encoding is detailed by Nanni and Lumini in their paper:
        L. Nanni and A. Lumini, “A new encoding technique for peptide classification,”
        Expert Syst. Appl., vol. 38, no. 4, pp. 3185–3191, 2011
        This function just receives 20aa letters, therefore preprocessing is required.

        :param n_jobs: number of CPU cores to be used. Default used is 4 CPU cores
        :return: Dataframe with the Fisher Transform encoding for each sequence.
        """
        path = self.path + '/adjuv_functions/features_functions/data/nlf.csv'
        encoding = pd.read_csv(path, index_col=0).to_dict()
        try:
            if isinstance(encoding, dict):
                with Parallel(n_jobs=n_jobs) as parallel:
                    res = parallel(delayed(seq_nlf_encoding)(seq, encoding) for seq in self.result[self.col])
                self.result = self.result.assign(nlf=res)
                return self.result
            else:
                raise Exception('The provided encoding is not valid')
        except:
            raise Exception('Sequences not preprocessed. Run sequence preprocessing')

    def get_zscale(self, n_jobs: int = 4):
        """
        This method encodes each amino acid of the sequence into a Z-scales. Each Z scale represent an amino-acid property:
        Z1: Lipophilicity
        Z2: Steric properties (Steric bulk/Polarizability)
        Z3: Electronic properties (Polarity / Charge)
        Z4 and Z5: They relate electronegativity, heat of formation, electrophilicity and hardness.

        :param n_jobs: number of CPU cores to be used. Default used is 4 CPU cores
        :return: Dataframe with the Z-scale encoding for each sequence.
        """
        encoding = zs
        try:
            if isinstance(encoding, dict):
                with Parallel(n_jobs=n_jobs) as parallel:
                    res = parallel(
                        delayed(seq_zscale_encoding)(seq, encoding) for seq in self.result[self.col])
                self.result = self.result.assign(zscale=res)
                return self.result
            else:
                raise Exception('The provided encoding is not valid')
        except:
            raise Exception('Sequences not preprocessed. Run sequence preprocessing')

    def get_esm(self, v_esm : str = 'esm2_150', batch : int = 4):
        """
        This method encodes each amino acid of the sequence by the ESM model transformer.
        More information available in: https://github.com/facebookresearch/esm

        :param v_esm: Version of the esm model
                    -'esm2_150' - model ESM2 with 150M parameters
                    -'esm2_650' - model ESM2 with 650M parameters
                    -'esm_1b' - model ESM-1b
        :param batch: number of sequences to process by batch (WARNING: for 16 gb of ram, 2 sequences per batch is
                        recommended)
        :return: Dataframe with the ESM encoding for each sequence.
        """
        import esm

        if v_esm == 'esm2_150':
            layer = 30
            model, alphabet = esm.pretrained.esm2_t30_150M_UR50D()
        elif v_esm == 'esm_1b':
            layer = 33
            model, alphabet = esm.pretrained.esm1b_t33_650M_UR50S()
        elif v_esm == 'esm2_650':
            layer = 33
            model, alphabet = esm.pretrained.esm2_t33_650M_UR50D()
        else:
            raise Exception('ESM model not available')

        alphabet = alphabet
        batch_converter = alphabet.get_batch_converter()
        model.eval()
        feature = []
        sequences =  self.result[self.col].tolist()
        sequences = [(str(seq_index), sequences[seq_index]) for seq_index in range(len(sequences))]
        for index in range(0, len(sequences), batch): #it has to run in batch, otherwise no 754 gb of ram not enought
            representations = {}
            print(str(index) + '/' + str(len(sequences)))
            if index + batch <= len(sequences):
                seq = sequences[index: index + batch]
            else:
                seq = sequences[index: len(sequences)]

            batch_labels, batch_strs, batch_tokens = batch_converter(seq)
            # Extract per-residue representations (on CPU)
            with torch.no_grad():
                results = model(batch_tokens, repr_layers=[layer], return_contacts=True)
            representations = results["representations"][layer]

            for seq_num in range(len(representations)):
                seq_emd = representations[seq_num].numpy()
                feature.append(seq_emd)

        self.result = self.result.assign(esm=feature)
        return self.result

    def get_protbert(self, batch : int =4):
        """
        This method encodes each amino acid of the sequence by the ProtBert model transformer.
        More information available in: https://github.com/agemagician/ProtTrans

        :param batch: number of sequences to process by batch (WARNING: for 16 gb of ram, 4 sequences per batch is
                        recommended)
        :return: Dataframe with the ProtBert encoding for each sequence.
        """
        import esm

        tokenizer = AutoTokenizer.from_pretrained("Rostlab/prot_bert", do_lower_case=False)
        model = AutoModel.from_pretrained("Rostlab/prot_bert")

        device = torch.device('cpu')
        model = model.eval()
        features = []
        for index in range(0, self.result[self.col].shape[0], batch):
            print(str(index) + '/' + str(self.result[self.col].shape[0]))
            if index + batch <= self.result[self.col].shape[0]:
                seq = self.result[self.col][index: index + batch].tolist()
            else:
                seq = self.result[self.col][index: self.result[self.col].shape[0]].tolist()

            ids = tokenizer.batch_encode_plus(seq, add_special_tokens=True)
            batch_tokens = torch.tensor(ids['input_ids']).to(device)
            attention_mask = torch.tensor(ids['attention_mask']).to(device)
            with torch.no_grad():
                embedding = model(input_ids=batch_tokens, attention_mask=attention_mask)[0]
            embedding = embedding.cpu().numpy()

            for seq_num in range(len(embedding)):
                seq_len = (attention_mask[seq_num] == 1).sum()
                seq_emd = embedding[seq_num][1:seq_len - 1]
                features.append(seq_emd)

        self.result = self.result.assign(protbert=features)
        return self.result

    def get_adaptable(self, list_of_functions: list, alphabet: str = "XARNDCEQGHILKMFPSTWYV", seq_len: int = 600,
                      padding_truncating='post',
                      blosum: str = 'blosum62', use_padded: bool = True, v_esm : str = 'esm2_150', batch : int = 4, n_jobs: int = 4):
        """
        It allows to run a selected set of encoding functions for the protein sequences.

        :param list_of_functions: A list with the fuctions to run.
        :param alphabet: The alphabet of aminoacids to be used.
        :param seq_len: The maximum length for all sequences. By default the length is 600 amino acids
        :param padding_truncating: 'pre' or 'post' pad either before or after each sequence, also removes values from sequences larger than seq_len. By default padding is after each sequence.
        :param blosum: blosum matrix to use either 'blosum62' or 'blosum50'. by default 'blosum62'
        :param use_padded: A bool argument, if true it uses the padded sequences for the one-hot, blosum and z-scale encoding.
        :param n_jobs: number of CPU cores to be used. Default used is 4 CPU cores
        :return: Dataframe with all encodings for each sequence.
        """
        import esm


        for function in list_of_functions:
            if function == 1:
                self.result = self.get_seq_pad(seq_len, alphabet, padding_truncating, n_jobs)
                if use_padded: self.col = 'padded_sequence'
            if function == 2: self.get_hot_encoded(alphabet, n_jobs)
            if function == 3: self.get_pad_and_hot_encoding(seq_len, alphabet, padding_truncating, n_jobs)
            if function == 4: self.get_blosum(blosum, n_jobs)
            if function == 5: self.get_nlf(n_jobs)
            if function == 6: self.get_zscale(n_jobs)
            if function == 7: self.get_esm(v_esm,batch)
            if function == 8: self.get_protbert(batch)
        return self.result


def seq_blosum_encoding(ProteinSequence: str, encoding: dict):
    """
    BLOSUM62 is a substitution matrix that specifies the similarity of one amino acid to another by means of a score.
    This score reflects the frequency of substitutions found from studying protein sequence conservation
    in large databases of related proteins.
    The number 62 refers to the percentage identity at which sequences are clustered in the analysis.
    I is possible to get blosum50 to get 50 % identity.
    Encoding a peptide this way means we provide the column from the blosum matrix corresponding to the amino acid
    at each position of the sequence. This produces 24*seqlen matrix.

    :param ProteinSequence: protein sequence to be encoded
    :param encoding: dict with the encoding. Keys are the aminoacid residues and values the encoding.
    :return: dict form with blosum encoding
    """
    return [list(encoding[i].values()) for i in ProteinSequence]


def seq_nlf_encoding(ProteinSequence: str, encoding: dict):
    """
    Method that takes many physicochemical properties and transforms them using a Fisher Transform (similar to a PCA)
    creating a smaller set of features that can describe the amino acid just as well.
    There are 19 transformed features.
    This method of encoding is detailed by Nanni and Lumini in their paper:
    L. Nanni and A. Lumini, “A new encoding technique for peptide classification,”
    Expert Syst. Appl., vol. 38, no. 4, pp. 3185–3191, 2011
    This function just receives 20aa letters, therefore preprocessing is required.

    :param ProteinSequence: protein sequence to be encoded
    :param encoding: dict with the encoding. Keys are the aminoacid residues and values the encoding.
    :return: Dict form with the Fisher Transform encoding for the sequence.
    """
    return [list(encoding[i].values()) for i in ProteinSequence]


def seq_zscale_encoding(ProteinSequence: str, encoding):
    """
    This method encodes each amino acid of the sequence into a Z-scales. Each Z scale represent an amino-acid property:
    Z1: Lipophilicity
    Z2: Steric properties (Steric bulk/Polarizability)
    Z3: Electronic properties (Polarity / Charge)
    Z4 and Z5: They relate electronegativity, heat of formation, electrophilicity and hardness.

    :param ProteinSequence: protein sequence to be encoded
    :param encoding: dict with the encoding. Keys are the aminoacid residues and values the encoding.
    :return: dict form with the Z-scale encoding for the sequence.
    """
    return [encoding[i] for i in ProteinSequence]


def seq_padding(ProteinSequence: str, char_to_int: dict, int_to_char: dict, seq_len: int,
                padding_truncating: str = 'post'):
    """
    This methods performs the padding of the sequence.

    :param ProteinSequence: protein sequence to be encoded
    :param char_to_int: dict with the encoding. Keys are the aminoacid residues and values the corresponding int.
    :param int_to_char: dict with the encoding. Keys are the corresponding int and the the values aminoacid residues.
    :param seq_len: The maximum length for all sequences. By default the length is 600 amino acids.
    :param padding_truncating: 'pre' or 'post' pad either before or after each sequence, also removes values from sequences larger than seq_len. By default padding is after each sequence.

    :return: Dict form with the padded sequences for the sequence.
    """
    integer_encoded = [[char_to_int[char] for char in ProteinSequence]]
    list_of_sequences_length = pad_sequences(integer_encoded, maxlen=seq_len, dtype='int32',
                                             padding=padding_truncating, truncating=padding_truncating, value=0)

    char_paded = [int_to_char[i] for i in list_of_sequences_length[0]]
    pad_aa = ''.join(char_paded)
    return pad_aa


def seq_hot_encoded(ProteinSequence: str, char_to_int: dict):
    """
    This methods performs the one-hot-encoding of the sequence.

    :param ProteinSequence: protein sequence to be encoded
    :param char_to_int: dict with the encoding. Keys are the aminoacid residues and values the corresponding int.

    :return: Dict form with the one-hot-encoding for the sequence.
    """
    integer_encoded = [char_to_int[char] for char in ProteinSequence]
    integer_encoded.append(len(char_to_int.keys()) - 1)
    hot_enc = to_categorical(integer_encoded)
    return hot_enc[:len(hot_enc) - 1]


def seq_padded_hot(ProteinSequence: str, char_to_int: dict, int_to_char: dict, seq_len: int,
                   padding_truncating: str = 'post'):
    """
    This methods performs the padding of the sequence and one-hot-encode the padded sequence.

    :param ProteinSequence: protein sequence to be encoded
    :param char_to_int: dict with the encoding. Keys are the aminoacid residues and values the corresponding int.
    :param int_to_char: dict with the encoding. Keys are the corresponding int and the the values aminoacid residues.
    :param seq_len: The maximum length for all sequences. By default the length is 600 amino acids.
    :param padding_truncating: 'pre' or 'post' pad either before or after each sequence, also removes values from sequences larger than seq_len. By default padding is after each sequence.

    :return: Dict form with the padded sequences and the one-hot-encoding for the sequence.
    """
    integer_encoded = [char_to_int[char] for char in ProteinSequence]
    list_of_sequences_length = pad_sequences([integer_encoded], maxlen=seq_len, dtype='int32',
                                             padding=padding_truncating, truncating=padding_truncating, value=0)

    integer_padded = np.append(list_of_sequences_length[0], [len(char_to_int.keys()) - 1])
    char_paded = [int_to_char[i] for i in integer_padded]
    pad_aa = ''.join(char_paded)

    hot_enc = to_categorical(integer_padded).tolist()

    return (hot_enc[:len(hot_enc) - 1], pad_aa[:len(pad_aa) - 1])


zs = {
    'A': [0.24, -2.32, 0.60, -0.14, 1.30],  # A
    'C': [0.84, -1.67, 3.71, 0.18, -2.65],  # C
    'D': [3.98, 0.93, 1.93, -2.46, 0.75],  # D
    'E': [3.11, 0.26, -0.11, -0.34, -0.25],  # E
    'F': [-4.22, 1.94, 1.06, 0.54, -0.62],  # F
    'G': [2.05, -4.06, 0.36, -0.82, -0.38],  # G
    'H': [2.47, 1.95, 0.26, 3.90, 0.09],  # H
    'I': [-3.89, -1.73, -1.71, -0.84, 0.26],  # I
    'K': [2.29, 0.89, -2.49, 1.49, 0.31],  # K
    'L': [-4.28, -1.30, -1.49, -0.72, 0.84],  # L
    'M': [-2.85, -0.22, 0.47, 1.94, -0.98],  # M
    'N': [3.05, 1.62, 1.04, -1.15, 1.61],  # N
    'P': [-1.66, 0.27, 1.84, 0.70, 2.00],  # P
    'Q': [1.75, 0.50, -1.44, -1.34, 0.66],  # Q
    'R': [3.52, 2.50, -3.50, 1.99, -0.17],  # R
    'S': [2.39, -1.07, 1.15, -1.39, 0.67],  # S
    'T': [0.75, -2.18, -1.12, -1.46, -0.40],  # T
    'V': [-2.59, -2.64, -1.54, -0.85, -0.02],  # V
    'W': [-4.36, 3.94, 0.59, 3.44, -1.59],  # W
    'Y': [-2.54, 2.44, 0.43, 0.04, -1.47],  # Y
    'X': [0.0, 0.0, 0.0, 0.0, 0.0]}
